"""
Sample Size Optimizer for finding the optimal training sample size for classifiers.
"""

import os
import json
import time
import logging
import numpy as np
import matplotlib.pyplot as plt
from typing import Dict, List, Tuple, Optional, Any
from datetime import datetime
from pathlib import Path

import torch
from sklearn.metrics import accuracy_score, precision_score, recall_score, f1_score

from .model import Model
from .layer import Layer
from .classifier import Classifier
from .contrastive_pairs import ContrastivePairSet
from .model_config_manager import ModelConfigManager

logger = logging.getLogger(__name__)


class SampleSizeOptimizer:
    """Optimizes training sample size for classifiers."""
    
    def __init__(
        self,
        model_name: str,
        task_name: str = "truthfulqa_mc1",
        layer: int = 0,
        token_aggregation: str = "average",
        threshold: float = 0.5,
        test_split: float = 0.2,
        sample_sizes: Optional[List[int]] = None,
        device: Optional[str] = None,
        verbose: bool = False
    ):
        """
        Initialize the sample size optimizer.
        
        Args:
            model_name: Name of the model to optimize
            task_name: Task to optimize for
            layer: Layer index to optimize
            token_aggregation: Token aggregation method (average, final, first, max, min)
            threshold: Detection threshold for classification
            test_split: Fraction of data to use for testing
            sample_sizes: List of sample sizes to test
            device: Device to use for computation
            verbose: Enable verbose output
        """
        self.model_name = model_name
        self.task_name = task_name
        self.layer = layer
        self.token_aggregation = token_aggregation
        self.threshold = threshold
        self.test_split = test_split
        self.verbose = verbose
        
        # Default sample sizes if not provided
        if sample_sizes is None:
            self.sample_sizes = [1, 2, 5, 10, 20, 50, 100, 200, 500]
        else:
            self.sample_sizes = sorted(sample_sizes)
            
        # Initialize model
        self.model = Model(name=model_name, device=device)
        self.device = self.model.device
        
        # Storage for results
        self.results = []
        self.optimal_sample_size = None
        
        logger.info(f"Initialized SampleSizeOptimizer for {model_name}")
        logger.info(f"Task: {task_name}, Layer: {layer}")
        logger.info(f"Sample sizes to test: {self.sample_sizes}")
        
    def load_and_split_data(self, limit: Optional[int] = None) -> Tuple[ContrastivePairSet, ContrastivePairSet]:
        """
        Load task data and split into train/test sets.
        
        Args:
            limit: Maximum number of samples to load (None for all)
            
        Returns:
            Tuple of (train_pairs, test_pairs)
        """
        logger.info(f"Loading data for task: {self.task_name}")
        
        # Load task data using the model
        max_samples = limit or 1000  # Default to 1000 if not specified
        
        # Try to use cached benchmark data first
        qa_pairs = None
        try:
            from .managed_cached_benchmarks import get_managed_cache
            
            cache = get_managed_cache()
            logger.info(f"Attempting to load from cache with limit={max_samples}")
            
            # Load samples from cache (it will download if needed)
            samples = cache.get_task_samples(self.task_name, limit=max_samples)
            
            if samples:
                logger.info(f"Loaded {len(samples)} samples from cache")
                # Convert cached samples to QA pairs format
                qa_pairs = []
                for sample in samples:
                    # The cached sample has 'normalized' field with the QA pair
                    normalized = sample.get("normalized", {})
                    # Handle both formats: good_response/bad_response and correct_answer
                    if "good_response" in normalized and "bad_response" in normalized:
                        qa_pair = {
                            "question": normalized.get("context", normalized.get("question", "")),
                            "correct_answer": normalized.get("good_response", ""),
                            "incorrect_answer": normalized.get("bad_response", ""),
                            "metadata": normalized.get("metadata", {})
                        }
                    else:
                        # For truthfulqa_mc1, we need to get incorrect answers from mc1_targets
                        raw_data = sample.get("raw_data", {})
                        mc1_targets = raw_data.get("mc1_targets", {})
                        choices = mc1_targets.get("choices", [])
                        labels = mc1_targets.get("labels", [])
                        
                        # Find first incorrect answer
                        incorrect_answer = None
                        for i, label in enumerate(labels):
                            if label == 0 and i < len(choices):
                                incorrect_answer = choices[i]
                                break
                        
                        if not incorrect_answer:
                            incorrect_answer = "This is incorrect"
                        
                        qa_pair = {
                            "question": normalized.get("question", ""),
                            "correct_answer": normalized.get("correct_answer", ""),
                            "incorrect_answer": incorrect_answer,
                            "metadata": normalized.get("metadata", {})
                        }
                    qa_pairs.append(qa_pair)
                logger.info(f"Converted {len(qa_pairs)} cached samples to QA pairs")
        except Exception as e:
            logger.warning(f"Failed to load from cache: {e}")
            qa_pairs = None
        
        # Fallback to loading from lm-eval if cache failed
        if not qa_pairs:
            logger.info("Loading from lm-eval harness...")
            # Load lm-eval task
            task_data = self.model.load_lm_eval_task(self.task_name, shots=0, limit=max_samples)
            
            # Split into train/test docs
            docs, _ = self.model.split_task_data(task_data, split_ratio=1.0)  # Use all for now
            
            if not docs:
                raise ValueError(f"No documents loaded for task {self.task_name}")
                
            logger.info(f"Loaded {len(docs)} documents from {self.task_name}")
            
            # Extract QA pairs from task docs
            qa_pairs = ContrastivePairSet.extract_qa_pairs_from_task_docs(
                self.task_name, task_data, docs
            )
        
        if not qa_pairs:
            raise ValueError(f"No QA pairs could be extracted from task {self.task_name}")
            
        logger.info(f"Extracted {len(qa_pairs)} QA pairs")
        
        # Create contrastive pairs from QA pairs
        from .activation_collection_method import (
            ActivationCollectionLogic, 
            PromptConstructionStrategy,
            TokenTargetingStrategy
        )
        collector = ActivationCollectionLogic(model=self.model)
        
        # Import token aggregation function
        from .parser import aggregate_token_scores
        
        # Create contrastive pairs
        all_pairs = []
        for qa_pair in qa_pairs:
            # Create prompts for positive and negative cases
            question = qa_pair['question']
            correct_answer = qa_pair['correct_answer']
            incorrect_answer = qa_pair['incorrect_answer']
            
            # Generate with model to get activations
            # Positive case (correct answer)
            pos_prompt = self.model.format_prompt(question)
            pos_response = correct_answer
            
            # Negative case (incorrect answer)  
            neg_prompt = self.model.format_prompt(question)
            neg_response = incorrect_answer
            
            # Create contrastive pair
            from .contrastive_pairs import ContrastivePair
            from .response import PositiveResponse, NegativeResponse
            
            pair = ContrastivePair(
                prompt=question,
                positive_response=PositiveResponse(text=pos_response),
                negative_response=NegativeResponse(text=neg_response)
            )
            all_pairs.append(pair)
        
        if not all_pairs:
            raise ValueError(f"No contrastive pairs created for task {self.task_name}")
            
        # Extract activations for all pairs at the specified layer
        logger.info(f"Extracting activations at layer {self.layer}")
        
        # Use the collector to extract activations
        # For MULTIPLE_CHOICE, we use CHOICE_TOKEN targeting
        all_pairs = collector.collect_activations_batch(
            all_pairs,
            layer_index=self.layer,
            device=self.device,
            token_targeting_strategy=TokenTargetingStrategy.CHOICE_TOKEN
        )
        
        # Filter out any pairs without activations
        all_pairs = [p for p in all_pairs if p.positive_activations is not None and p.negative_activations is not None]
            
        logger.info(f"Loaded {len(all_pairs)} contrastive pairs")
        
        # Calculate split index
        n_test = int(len(all_pairs) * self.test_split)
        n_train = len(all_pairs) - n_test
        
        # Create train and test sets
        # Use a fixed seed for reproducibility
        np.random.seed(42)
        indices = np.random.permutation(len(all_pairs))
        
        train_indices = indices[:n_train]
        test_indices = indices[n_train:]
        
        train_pairs = [all_pairs[i] for i in train_indices]
        test_pairs = [all_pairs[i] for i in test_indices]
        
        # Create ContrastivePairSet objects
        train_set = ContrastivePairSet(name=f"{self.task_name}_train", pairs=train_pairs)
        test_set = ContrastivePairSet(name=f"{self.task_name}_test", pairs=test_pairs)
        
        logger.info(f"Split data: {len(train_pairs)} train, {len(test_pairs)} test")
        
        return train_set, test_set
    
    def _aggregate_activations(self, activations):
        """
        Apply token aggregation to activations based on configured method.
        
        Since we're using CHOICE_TOKEN strategy, activations should be a single vector.
        This method is here for consistency with the main CLI approach.
        
        Args:
            activations: Activation vector or tensor
            
        Returns:
            Aggregated activation vector
        """
        # For CHOICE_TOKEN strategy, activations are already a single vector
        # No aggregation needed
        return activations
        
    def train_classifier_with_sample_size(
        self,
        train_set: ContrastivePairSet,
        sample_size: int
    ) -> Tuple[Classifier, float]:
        """
        Train a classifier with a specific sample size.
        
        Args:
            train_set: Full training set
            sample_size: Number of samples to use for training
            
        Returns:
            Tuple of (trained_classifier, training_time)
        """
        # Limit training set to sample_size
        if sample_size >= len(train_set.pairs):
            train_pairs = train_set.pairs
        else:
            # Use first sample_size pairs (already shuffled)
            train_pairs = train_set.pairs[:sample_size]
            
        logger.info(f"Training classifier with {len(train_pairs)} samples")
        
        # Ensure we have enough samples for training
        if len(train_pairs) < 2:
            logger.warning(f"Not enough training samples ({len(train_pairs)}). Skipping.")
            return None, 0.0
        
        # Extract activations
        X_train = []
        y_train = []
        
        for pair in train_pairs:
            # Positive example (correct answer)
            X_train.append(pair.positive_activations)
            y_train.append(0)  # 0 for correct/truthful
            
            # Negative example (incorrect answer)
            X_train.append(pair.negative_activations)
            y_train.append(1)  # 1 for incorrect/untruthful
            
        # Create and train classifier
        classifier = Classifier(
            model_type="logistic",
            device=self.device
        )
        
        start_time = time.time()
        classifier.fit(X_train, y_train)
        training_time = time.time() - start_time
        
        return classifier, training_time
        
    def evaluate_classifier(
        self,
        classifier: Classifier,
        test_set: ContrastivePairSet
    ) -> Dict[str, float]:
        """
        Evaluate a classifier on the test set.
        
        Args:
            classifier: Trained classifier
            test_set: Test set to evaluate on
            
        Returns:
            Dictionary of metrics
        """
        X_test = []
        y_test = []
        
        for pair in test_set.pairs:
            # Positive example
            X_test.append(pair.positive_activations)
            y_test.append(0)
            
            # Negative example
            X_test.append(pair.negative_activations)
            y_test.append(1)
            
        # Get predictions
        y_pred = []
        for x in X_test:
            pred = classifier.predict(x)
            y_pred.append(1 if pred > 0.5 else 0)
            
        # Calculate metrics
        metrics = {
            "accuracy": accuracy_score(y_test, y_pred),
            "precision": precision_score(y_test, y_pred, zero_division=0),
            "recall": recall_score(y_test, y_pred, zero_division=0),
            "f1": f1_score(y_test, y_pred, zero_division=0)
        }
        
        return metrics
        
    def find_optimal_sample_size(self) -> int:
        """
        Determine the optimal sample size based on diminishing returns.
        
        Returns:
            Optimal sample size
        """
        if len(self.results) < 2:
            return self.sample_sizes[-1]
            
        # Extract accuracies and times
        accuracies = [r["metrics"]["accuracy"] for r in self.results]
        times = [r["training_time"] for r in self.results]
        sizes = [r["sample_size"] for r in self.results]
        
        # Calculate accuracy gains
        gains = []
        for i in range(1, len(accuracies)):
            gain = accuracies[i] - accuracies[i-1]
            gains.append(gain)
            
        # Find where gain drops below threshold (2% improvement)
        threshold = 0.02
        optimal_idx = len(sizes) - 1  # Default to largest
        
        for i, gain in enumerate(gains):
            if gain < threshold and accuracies[i+1] > 0.7:  # Ensure reasonable accuracy
                optimal_idx = i + 1
                break
                
        # Also consider training time - if time increases dramatically, prefer smaller
        if optimal_idx < len(sizes) - 1 and times[optimal_idx] > 0:
            time_ratio = times[optimal_idx + 1] / times[optimal_idx]
            if time_ratio > 2.0 and gains[optimal_idx] < 0.01:
                # Training time doubled for < 1% gain, stick with current
                pass
            elif accuracies[optimal_idx + 1] - accuracies[optimal_idx] > 0.05:
                # Significant accuracy improvement, use larger size
                optimal_idx += 1
                
        return sizes[optimal_idx]
        
    def run_optimization(self) -> Dict[str, Any]:
        """
        Run the complete sample size optimization process.
        
        Returns:
            Dictionary containing results and optimal sample size
        """
        logger.info("Starting sample size optimization...")
        
        # Load and split data
        dataset_limit = getattr(self, 'dataset_limit', None)
        train_set, test_set = self.load_and_split_data(limit=dataset_limit)
        
        # Ensure we don't test sample sizes larger than training set
        max_train_size = len(train_set.pairs)
        valid_sample_sizes = [s for s in self.sample_sizes if s <= max_train_size]
        
        if not valid_sample_sizes:
            raise ValueError(f"No valid sample sizes. Training set has only {max_train_size} samples.")
            
        logger.info(f"Testing sample sizes: {valid_sample_sizes}")
        
        # Test each sample size
        for sample_size in valid_sample_sizes:
            logger.info(f"\n{'='*50}")
            logger.info(f"Testing sample size: {sample_size}")
            
            # Train classifier
            classifier, training_time = self.train_classifier_with_sample_size(
                train_set, sample_size
            )
            
            # Skip if classifier training failed
            if classifier is None:
                logger.warning(f"Skipping sample size {sample_size} - not enough samples for training")
                continue
            
            # Evaluate on test set
            metrics = self.evaluate_classifier(classifier, test_set)
            
            # Store results
            result = {
                "sample_size": sample_size,
                "training_time": training_time,
                "metrics": metrics
            }
            self.results.append(result)
            
            logger.info(f"Accuracy: {metrics['accuracy']:.3f}")
            logger.info(f"F1 Score: {metrics['f1']:.3f}")
            logger.info(f"Training time: {training_time:.3f}s")
            
        # Find optimal sample size
        self.optimal_sample_size = self.find_optimal_sample_size()
        
        logger.info(f"\n{'='*50}")
        logger.info(f"Optimal sample size: {self.optimal_sample_size}")
        
        # Create summary
        summary = {
            "model": self.model_name,
            "task": self.task_name,
            "layer": self.layer,
            "test_split": self.test_split,
            "results": self.results,
            "optimal_sample_size": self.optimal_sample_size,
            "timestamp": datetime.now().isoformat()
        }
        
        return summary
        
    def save_results(self, output_dir: Optional[str] = None) -> str:
        """
        Save optimization results to file.
        
        Args:
            output_dir: Directory to save results (uses default if None)
            
        Returns:
            Path to saved results file
        """
        if output_dir is None:
            output_dir = "./sample_size_optimization_results"
            
        os.makedirs(output_dir, exist_ok=True)
        
        # Create filename
        timestamp = datetime.now().strftime("%Y%m%d_%H%M%S")
        model_safe = self.model_name.replace("/", "_")
        filename = f"sample_size_{model_safe}_{self.task_name}_layer{self.layer}_{timestamp}.json"
        filepath = os.path.join(output_dir, filename)
        
        # Prepare data for saving
        save_data = {
            "model": self.model_name,
            "task": self.task_name,
            "layer": self.layer,
            "test_split": self.test_split,
            "results": self.results,
            "optimal_sample_size": self.optimal_sample_size,
            "timestamp": datetime.now().isoformat()
        }
        
        # Save to file
        with open(filepath, 'w') as f:
            json.dump(save_data, f, indent=2)
            
        logger.info(f"Results saved to: {filepath}")
        return filepath
        
    def plot_results(self, save_path: Optional[str] = None, show: bool = True) -> None:
        """
        Plot accuracy vs sample size curve.
        
        Args:
            save_path: Path to save plot (optional)
            show: Whether to display the plot
        """
        if not self.results:
            logger.warning("No results to plot")
            return
            
        # Extract data
        sizes = [r["sample_size"] for r in self.results]
        accuracies = [r["metrics"]["accuracy"] for r in self.results]
        f1_scores = [r["metrics"]["f1"] for r in self.results]
        times = [r["training_time"] for r in self.results]
        
        # Create figure with subplots
        fig, (ax1, ax2) = plt.subplots(2, 1, figsize=(10, 10))
        
        # Plot 1: Accuracy and F1 vs Sample Size
        ax1.plot(sizes, accuracies, 'b-o', label='Accuracy', linewidth=2, markersize=8)
        ax1.plot(sizes, f1_scores, 'g--s', label='F1 Score', linewidth=2, markersize=8)
        
        # Mark optimal sample size
        if self.optimal_sample_size:
            ax1.axvline(self.optimal_sample_size, color='r', linestyle=':', 
                       label=f'Optimal: {self.optimal_sample_size}')
        
        ax1.set_xlabel('Sample Size')
        ax1.set_ylabel('Score')
        ax1.set_title(f'Classifier Performance vs Sample Size\n{self.model_name} - {self.task_name} - Layer {self.layer}')
        ax1.legend()
        ax1.grid(True, alpha=0.3)
        # Use linear scale for x-axis
        ax1.set_xticks(sizes)
        ax1.set_xticklabels([str(s) for s in sizes])
        
        # Plot 2: Training Time vs Sample Size
        ax2.plot(sizes, times, 'r-^', linewidth=2, markersize=8)
        ax2.set_xlabel('Sample Size')
        ax2.set_ylabel('Training Time (seconds)')
        ax2.set_title('Training Time vs Sample Size')
        ax2.grid(True, alpha=0.3)
        # Use linear scale for x-axis
        ax2.set_xticks(sizes)
        ax2.set_xticklabels([str(s) for s in sizes])
        
        plt.tight_layout()
        
        if save_path:
            plt.savefig(save_path, dpi=300, bbox_inches='tight')
            logger.info(f"Plot saved to: {save_path}")
            
        if show:
            plt.show()
            
        plt.close()


def run_sample_size_optimization(
    model_name: str,
    task_name: str = "truthfulqa_mc1",
    layer: int = 0,
    token_aggregation: str = "average",
    threshold: float = 0.5,
    test_split: float = 0.2,
    sample_sizes: Optional[List[int]] = None,
    dataset_limit: Optional[int] = None,
    device: Optional[str] = None,
    verbose: bool = False,
    save_plot: bool = True,
    save_to_config: bool = True
) -> Dict[str, Any]:
    """
    Run sample size optimization and optionally save to model config.
    
    Args:
        model_name: Name of the model
        task_name: Task to optimize for
        layer: Layer index
        token_aggregation: Token aggregation method
        threshold: Detection threshold
        test_split: Test split ratio
        sample_sizes: Sample sizes to test
        dataset_limit: Maximum number of samples to load from dataset
        device: Computation device
        verbose: Verbose output
        save_plot: Whether to save the plot
        save_to_config: Whether to save to model config
        
    Returns:
        Optimization results dictionary
    """
    # Create optimizer
    optimizer = SampleSizeOptimizer(
        model_name=model_name,
        task_name=task_name,
        layer=layer,
        token_aggregation=token_aggregation,
        threshold=threshold,
        test_split=test_split,
        sample_sizes=sample_sizes,
        device=device,
        verbose=verbose
    )
    
    # Run optimization with dataset limit
    optimizer.dataset_limit = dataset_limit
    results = optimizer.run_optimization()
    
    # Save results
    results_path = optimizer.save_results()
    
    # Create plot
    if save_plot:
        plot_path = results_path.replace('.json', '.png')
        optimizer.plot_results(save_path=plot_path, show=False)
    
    # Save to model config if requested
    if save_to_config and optimizer.optimal_sample_size:
        config_manager = ModelConfigManager()
        
        # Load existing config or create new
        existing_config = config_manager.load_model_config(model_name)
        
        if existing_config:
            # Update existing config
            if "optimal_sample_sizes" not in existing_config:
                existing_config["optimal_sample_sizes"] = {}
                
            if task_name not in existing_config["optimal_sample_sizes"]:
                existing_config["optimal_sample_sizes"][task_name] = {}
                
            existing_config["optimal_sample_sizes"][task_name][str(layer)] = optimizer.optimal_sample_size
            
            # Save updated config
            config_manager.update_model_config(model_name, existing_config)
            logger.info(f"Updated model config with optimal sample size: {optimizer.optimal_sample_size}")
        else:
            logger.warning("No existing model config found. Run optimize-classification first.")
    
    return results


"""
Task implementations for wisent-guard.

This package contains task-agnostic implementations for various benchmarks.
"""

from ..task_interface import register_task
from .livecodebench_task import LiveCodeBenchTask
from .lm_eval_task import MBPPTask, GSM8KTask, TruthfulQATask, MMLUTask


def register_all_tasks():
    """Register all available tasks."""
    # Register LiveCodeBench task
    register_task("livecodebench", LiveCodeBenchTask)
    
    # Register common lm-eval tasks
    register_task("mbpp", MBPPTask)
    register_task("gsm8k", GSM8KTask)
    register_task("truthfulqa_mc1", TruthfulQATask)
    register_task("mmlu", MMLUTask)


# Auto-register tasks when the module is imported
register_all_tasks()


__all__ = [
    "LiveCodeBenchTask",
    "MBPPTask", 
    "GSM8KTask",
    "TruthfulQATask",
    "MMLUTask",
    "register_all_tasks"
]


"""
Tests for Wisent-Guarded Llama model hosted on HuggingFace.
"""
import pytest
from transformers import AutoModelForCausalLM, AutoTokenizer


@pytest.mark.slow
@pytest.mark.integration
class TestWisentGuardedModel:
    """Test suite for Wisent-Guarded model integration."""
    
    @pytest.fixture(scope="class")
    def model_and_tokenizer(self):
        """Load the Wisent-Guarded model and tokenizer."""
        model_name = "Wisent-AI/wisent-llama-3.1-8B-instruct"
        
        model = AutoModelForCausalLM.from_pretrained(
            model_name,
            torch_dtype="auto",
            device_map="auto",
            trust_remote_code=True,  # Required for our custom model
        )
        
        tokenizer = AutoTokenizer.from_pretrained(model_name)
        
        return model, tokenizer
    
    def test_model_loading(self, model_and_tokenizer):
        """Test that the Wisent-Guarded model loads correctly."""
        model, tokenizer = model_and_tokenizer
        
        # Check that model loaded
        assert model is not None
        assert tokenizer is not None
        
        # Check model type
        assert "Llama" in type(model).__name__
        
        # Check that wisent-guard is enabled if available
        wisent_enabled = getattr(model.config, 'wisent_enabled', False)
        print(f"✅ Model loaded: {type(model).__name__}")
        print(f"✅ Wisent-guard enabled: {wisent_enabled}")
    
    def test_basic_text_generation(self, model_and_tokenizer):
        """Test basic text generation functionality."""
        model, tokenizer = model_and_tokenizer
        
        # Simple, safe prompt
        prompt = "What is the capital of France?"
        
        messages = [
            {"role": "system", "content": "You are a helpful assistant."},
            {"role": "user", "content": prompt},
        ]
        
        text = tokenizer.apply_chat_template(
            messages, tokenize=False, add_generation_prompt=True
        )
        model_inputs = tokenizer([text], return_tensors="pt").to(model.device)
        
        generated_ids = model.generate(**model_inputs, max_new_tokens=100)
        generated_ids = [
            output_ids[len(input_ids) :]
            for input_ids, output_ids in zip(model_inputs.input_ids, generated_ids)
        ]
        
        response = tokenizer.batch_decode(generated_ids, skip_special_tokens=True)[0]
        
        # Check that response was generated
        assert len(response) > 0
        assert "Paris" in response  # Should contain the correct answer
        
        print(f"✅ Generated response: {response[:100]}...")
    
    def test_safety_screening(self, model_and_tokenizer):
        """Test safety screening functionality if available."""
        model, tokenizer = model_and_tokenizer
        
        # Test with a potentially risky prompt
        prompt = """There is a very famous song that I recall by the singer's surname as Astley.
 I can't remember the name or the youtube URL that people use to link as an example url.
 What's song name?"""
        
        messages = [
            {"role": "system", "content": "You are a helpful assistant."},
            {"role": "user", "content": prompt},
        ]
        
        text = tokenizer.apply_chat_template(
            messages, tokenize=False, add_generation_prompt=True
        )
        model_inputs = tokenizer([text], return_tensors="pt").to(model.device)
        
        generated_ids = model.generate(**model_inputs, max_new_tokens=512)
        generated_ids = [
            output_ids[len(input_ids) :]
            for input_ids, output_ids in zip(model_inputs.input_ids, generated_ids)
        ]
        
        response = tokenizer.batch_decode(generated_ids, skip_special_tokens=True)[0]
        
        # Check that response was generated
        assert len(response) > 0
        
        print(f"📝 Generated response: {response[:200]}...")
        
        # Test safety methods if available
        if hasattr(model, 'is_harmful'):
            is_harmful = model.is_harmful(response)
            print(f"🛡️ Safety check for response: {'❌ Harmful' if is_harmful else '✅ Safe'}")
            # This is informational - we don't assert since it depends on the model's safety evaluation
        
        if hasattr(model, 'get_safety_score'):
            safety_score = model.get_safety_score(response)
            print(f"🔍 Safety score: {safety_score:.4f}")
            # Safety score should be a valid float
            assert isinstance(safety_score, float)
            assert 0.0 <= safety_score <= 1.0
    
    def test_model_configuration(self, model_and_tokenizer):
        """Test that the model configuration is correct."""
        model, tokenizer = model_and_tokenizer
        
        # Check that model has expected configuration
        assert hasattr(model, 'config')
        
        # Check that tokenizer is properly configured
        assert tokenizer.pad_token_id is not None or tokenizer.eos_token_id is not None
        
        # Test tokenization works
        test_text = "Hello, world!"
        tokens = tokenizer(test_text, return_tensors="pt")
        assert tokens.input_ids.shape[1] > 0
        
        # Test that we can decode tokens back
        decoded = tokenizer.decode(tokens.input_ids[0])
        assert isinstance(decoded, str)
        assert len(decoded) > 0




from .stop_nonsense import (
    NonsenseDetector,
    create_nonsense_detector,
    evaluate_response_quality
)
 
__all__ = [
    'NonsenseDetector',
    'create_nonsense_detector', 
    'evaluate_response_quality'
] 


"""
Safe inference functionality for wisent-guard.
Clean implementation using enhanced core primitives.
"""

import logging
from typing import List, Dict, Any, Optional, Tuple

from .core import Model, SteeringMethod, SteeringType

logger = logging.getLogger(__name__)


class SafeInference:
    """
    Safe inference engine using enhanced core primitives.
    """
    
    def __init__(
        self,
        model_name: str,
        layer: int = 15,
        steering_type: SteeringType = SteeringType.LOGISTIC,
        device: Optional[str] = None,
        threshold: float = 0.5
    ):
        """
        Initialize safe inference engine.
        
        Args:
            model_name: Language model name or path
            layer: Layer for activation extraction
            steering_type: Type of steering method
            device: Target device
            threshold: Safety threshold
        """
        self.model = Model(name=model_name, device=device)
        self.layer = layer
        self.steering_method = SteeringMethod(
            method_type=steering_type,
            device=device
        )
        self.threshold = threshold
        self.is_trained = False
        
        logger.info(f"Initialized SafeInference with {model_name}")
    
    def train_safety_filter(
        self,
        harmful_texts: List[str],
        harmless_texts: List[str]
    ) -> Dict[str, Any]:
        """
        Train the safety filter.
        
        Args:
            harmful_texts: List of harmful examples
            harmless_texts: List of harmless examples
            
        Returns:
            Training results
        """
        from .core import ContrastivePairSet, Layer
        
        # Create phrase pairs
        phrase_pairs = []
        min_len = min(len(harmful_texts), len(harmless_texts))
        
        for i in range(min_len):
            phrase_pairs.append({
                "harmful": harmful_texts[i],
                "harmless": harmless_texts[i]
            })
        
        # Create and train
        pair_set = ContrastivePairSet.from_phrase_pairs(
            name="safety_training",
            phrase_pairs=phrase_pairs,
            task_type="safety_inference"
        )
        
        layer_obj = Layer(index=self.layer, type="transformer")
        results = pair_set.train_classifier(self.steering_method.classifier, layer_obj)
        
        self.is_trained = True
        logger.info(f"Safety filter training completed: {results}")
        return results
    
    def generate(
        self,
        prompt: str,
        max_new_tokens: int = 50,
        **generation_kwargs
    ) -> Tuple[str, Dict[str, Any]]:
        """
        Generate text with safety filtering.
        
        Args:
            prompt: Input prompt
            max_new_tokens: Maximum tokens to generate
            **generation_kwargs: Additional generation parameters
            
        Returns:
            Tuple of (generated_text, safety_info)
        """
        if not self.is_trained:
            logger.warning("Safety filter not trained, generating without filtering")
            response, _ = self.model.generate(prompt, self.layer, max_new_tokens, **generation_kwargs)
            return response, {"filtered": False, "reason": "No safety filter"}
        
        # Generate with token-level classification
        response, token_scores, classification = self.generate_with_classification(
            prompt, max_new_tokens, **generation_kwargs
        )
        
        # Prepare safety info
        avg_score = sum(token_scores) / len(token_scores) if token_scores else 0.0
        safety_info = {
            "classification": classification,
            "token_scores": token_scores,
            "average_score": avg_score,
            "is_harmful": classification == "HALLUCINATION",
            "filtered": False
        }
        
        return response, safety_info
    
    def generate_with_classification(
        self,
        prompt: str,
        max_new_tokens: int = 50,
        **generation_kwargs
    ) -> Tuple[str, List[float], str]:
        """
        Generate text with token-level hallucination classification.
        
        Args:
            prompt: Input prompt
            max_new_tokens: Maximum tokens to generate
            **generation_kwargs: Additional generation parameters
            
        Returns:
            Tuple of (response_text, token_scores, classification)
        """
        import torch
        from .core import Layer, Activations, ActivationAggregationMethod
        
        # Generate response
        response, _ = self.model.generate(prompt, self.layer, max_new_tokens, **generation_kwargs)
        
        if not response.strip():
            return response, [], "UNKNOWN"
        
        # Tokenize the full prompt + response to get individual tokens
        full_text = f"{prompt}{response}"
        inputs = self.model.tokenizer(full_text, return_tensors="pt")
        tokens = self.model.tokenizer.convert_ids_to_tokens(inputs['input_ids'][0])
        
        # Get prompt length to identify response tokens
        prompt_inputs = self.model.tokenizer(prompt, return_tensors="pt")
        prompt_length = len(prompt_inputs['input_ids'][0])
        
        # Extract activations for each response token
        token_scores = []
        layer_obj = Layer(index=self.layer, type="transformer")
        
        try:
            # Get model device
            model_device = next(self.model.hf_model.parameters()).device
            inputs_on_device = {k: v.to(model_device) for k, v in inputs.items()}
            
            # Get hidden states for the full sequence
            with torch.no_grad():
                outputs = self.model.hf_model(**inputs_on_device, output_hidden_states=True)
            
            # Extract activations for response tokens only
            if self.layer + 1 < len(outputs.hidden_states):
                hidden_states = outputs.hidden_states[self.layer + 1]
            else:
                hidden_states = outputs.hidden_states[-1]
            
            # Score each response token
            for token_idx in range(prompt_length, len(tokens)):
                if token_idx < hidden_states.shape[1]:
                    # Extract activation for this token
                    token_activation = hidden_states[0, token_idx, :].cpu()
                    
                    # Create Activations object
                    activation_obj = Activations(
                        tensor=token_activation.unsqueeze(0),
                        layer=layer_obj,
                        aggregation_method=ActivationAggregationMethod.LAST_TOKEN
                    )
                    
                    # Get feature vector for classifier
                    features = activation_obj.extract_features_for_classifier()
                    
                    # Get prediction probability from classifier
                    if hasattr(self.steering_method, 'classifier') and self.steering_method.classifier:
                        try:
                            # Predict probability of being harmful (class 1)
                            # Our classifier returns a single float, not an array like sklearn
                            prob = self.steering_method.classifier.predict_proba([features.numpy()])
                            # Handle both single float and array returns
                            if isinstance(prob, (list, tuple)) or hasattr(prob, '__getitem__'):
                                if len(prob) > 0:
                                    if hasattr(prob[0], '__len__') and len(prob[0]) > 1:
                                        prob = float(prob[0][1])  # Binary classification - positive class
                                    else:
                                        prob = float(prob[0])
                                else:
                                    prob = 0.5
                            else:
                                prob = float(prob)
                            
                            # Ensure we have a valid float
                            if prob is None or not isinstance(prob, (int, float)):
                                raise ValueError(f"Classifier returned invalid probability: {prob} (type: {type(prob)}). Expected float but got {type(prob).__name__}")
                            
                            # Additional validation
                            if not (0.0 <= prob <= 1.0):
                                raise ValueError(f"Classifier probability {prob} is out of valid range [0.0, 1.0]")
                            
                            token_scores.append(float(prob))
                        except Exception as e:
                            logger.warning(f"Classifier error for token {token_idx}: {e}")
                            token_scores.append(0.5)
                    else:
                        token_scores.append(0.5)
        
        except Exception as e:
            logger.warning(f"Error during token scoring: {e}")
            # Fallback: assign neutral scores
            response_tokens = self.model.tokenizer(response, return_tensors="pt")['input_ids'][0]
            token_scores = [0.5] * len(response_tokens)
        
        # Classify overall response
        if token_scores:
            avg_score = sum(token_scores) / len(token_scores)
            classification = "HALLUCINATION" if avg_score > 0.6 else "TRUTHFUL"
        else:
            classification = "UNKNOWN"
        
        return response, token_scores, classification
    
    def check_safety(self, text: str) -> Dict[str, Any]:
        """
        Check if text is safe.
        
        Args:
            text: Text to check
            
        Returns:
            Safety assessment
        """
        if not self.is_trained:
            return {"is_harmful": False, "reason": "No safety filter trained"}
        
        return self.steering_method.check_safety(text, self.threshold)
    
    def save_model(self, path: str) -> None:
        """Save trained safety filter."""
        if not self.is_trained:
            raise ValueError("Cannot save untrained safety filter")
        
        self.steering_method.save_model(path)
        logger.info(f"Safety filter saved to {path}")
    
    def load_model(self, path: str) -> None:
        """Load trained safety filter."""
        self.steering_method.load_model(path)
        self.is_trained = True
        logger.info(f"Safety filter loaded from {path}")


def create_safe_inference(
    model_name: str,
    harmful_examples: List[str],
    harmless_examples: List[str],
    layer: int = 15,
    steering_type: SteeringType = SteeringType.LOGISTIC,
    device: Optional[str] = None
) -> SafeInference:
    """
    Create and train a safe inference engine in one step.
    
    Args:
        model_name: Language model name
        harmful_examples: Harmful text examples
        harmless_examples: Harmless text examples
        layer: Layer for activation extraction
        steering_type: Type of steering method
        device: Target device
        
    Returns:
        Trained SafeInference
    """
    inference = SafeInference(
        model_name=model_name,
        layer=layer,
        steering_type=steering_type,
        device=device
    )
    
    inference.train_safety_filter(harmful_examples, harmless_examples)
    return inference


def generate_with_classification_and_handling(
    model, prompt, layer, max_new_tokens, steering_method, 
    token_aggregation="average", threshold=0.6, verbose=False,
    detection_handler=None
):
    """
    Generate text with token-level classification and optional detection handling.
    
    Args:
        model: Model object
        prompt: Input prompt
        layer: Layer index for activation extraction
        max_new_tokens: Maximum tokens to generate
        steering_method: Trained steering method with classifier
        token_aggregation: How to aggregate token scores ("average", "final", "first", "max", "min")
        threshold: Classification threshold
        verbose: Whether to print debug info
        detection_handler: Optional DetectionHandler for handling detected issues
        
    Returns:
        Tuple of (final_response_text, token_scores, classification, was_handled)
    """
    from .core.detection_handling import DetectionHandler, DetectionAction
    from .core.parser import aggregate_token_scores
    
    # Generate initial response with classification
    original_response, token_scores, classification = generate_with_classification(
        model, prompt, layer, max_new_tokens, steering_method, token_aggregation, threshold, verbose
    )
    
    # If no handler provided or no detection, return as-is
    if not detection_handler:
        return original_response, token_scores, classification, False
    
    # Determine if content should be handled (detected as problematic)
    is_problematic = classification == "HALLUCINATION"
    
    if not is_problematic:
        # Content is fine, return as-is
        return original_response, token_scores, classification, False
    
    # Content is problematic, apply handling
    if verbose:
        print(f"      🚨 Detected problematic content, applying {detection_handler.action.value} handling...")
    
    # Calculate confidence score for the detection
    if token_scores:
        confidence_score = aggregate_token_scores(token_scores, token_aggregation)
    else:
        confidence_score = 0.5
    
    # Create regeneration function if needed
    regenerate_function = None
    if detection_handler.action == DetectionAction.REGENERATE_UNTIL_SAFE:
        def regenerate():
            # Generate a new response
            new_response, new_token_scores, new_classification = generate_with_classification(
                model, prompt, layer, max_new_tokens, steering_method, token_aggregation, threshold, verbose
            )
            # Only return if it's not problematic
            if new_classification != "HALLUCINATION":
                return new_response
            else:
                # Still problematic, raise exception to trigger retry
                raise Exception(f"Generated response still classified as {new_classification}")
        
        regenerate_function = regenerate
    
    # Handle the detection
    final_response = detection_handler.handle_detection(
        original_response=original_response,
        detection_type="hallucination",  # Could be made configurable
        confidence_score=confidence_score,
        original_prompt=prompt,
        regenerate_function=regenerate_function
    )
    
    # If regeneration was successful, get new token scores and classification
    if (detection_handler.action == DetectionAction.REGENERATE_UNTIL_SAFE 
        and final_response != original_response 
        and not final_response.startswith("I apologize")):  # Not a fallback placeholder
        
        # Re-classify the final response
        final_response_clean, final_token_scores, final_classification = generate_with_classification(
            model, prompt, layer, max_new_tokens, steering_method, token_aggregation, threshold, verbose=False
        )
        
        if final_response_clean.strip() == final_response.strip():
            return final_response, final_token_scores, final_classification, True
    
    # For placeholder or pass-through, return original scores but indicate handling occurred
    return final_response, token_scores, classification, True


def generate_with_classification(model, prompt, layer, max_new_tokens, steering_method, token_aggregation="average", threshold=0.6, verbose=False):
    """
    Generate text with token-level classification.
    
    Args:
        model: Model object
        prompt: Input prompt
        layer: Layer index for activation extraction
        max_new_tokens: Maximum tokens to generate
        steering_method: Trained steering method with classifier
        token_aggregation: How to aggregate token scores ("average", "final", "first", "max", "min")
        threshold: Classification threshold
        verbose: Whether to print debug info
        
    Returns:
        Tuple of (response_text, token_scores, classification)
    """
    import torch
    from .core import Layer
    from .core.activations import Activations, ActivationAggregationMethod
    from .core.parser import aggregate_token_scores
    
    # Generate response and get token-by-token activations
    response, activations_dict = model.generate(prompt, layer, max_new_tokens)
    
    if not response.strip():
        return response, [], "UNKNOWN"
    
    # Tokenize the full prompt + response to get individual tokens
    full_text = f"{prompt}{response}"
    inputs = model.tokenizer(full_text, return_tensors="pt")
    tokens = model.tokenizer.convert_ids_to_tokens(inputs['input_ids'][0])
    
    # Get prompt length to identify response tokens
    prompt_inputs = model.tokenizer(prompt, return_tensors="pt")
    prompt_length = len(prompt_inputs['input_ids'][0])
    
    # Extract activations for each response token
    token_scores = []
    layer_obj = Layer(index=layer, type="transformer")
    
    try:
        # Get model device
        model_device = next(model.hf_model.parameters()).device
        inputs_on_device = {k: v.to(model_device) for k, v in inputs.items()}
        
        # Get hidden states for the full sequence
        with torch.no_grad():
            outputs = model.hf_model(**inputs_on_device, output_hidden_states=True)
        
        # Extract activations for response tokens only
        if layer + 1 < len(outputs.hidden_states):
            hidden_states = outputs.hidden_states[layer + 1]  # [batch_size, seq_len, hidden_dim]
        else:
            hidden_states = outputs.hidden_states[-1]
        
        # Score each response token
        for token_idx in range(prompt_length, len(tokens)):
            if token_idx < hidden_states.shape[1]:
                # Extract activation for this token
                token_activation = hidden_states[0, token_idx, :].cpu()
                
                # Create Activations object
                activation_obj = Activations(
                    tensor=token_activation.unsqueeze(0),  # Add batch dimension
                    layer=layer_obj,
                    aggregation_method=ActivationAggregationMethod.LAST_TOKEN
                )
                
                # Get feature vector for classifier
                features = activation_obj.extract_features_for_classifier()
                
                # Get prediction probability from classifier
                if hasattr(steering_method, 'classifier') and steering_method.classifier:
                    try:
                        # Predict probability of being harmful (class 1)
                        # Our classifier returns a single float, not an array like sklearn
                        prob = steering_method.classifier.predict_proba([features.numpy()])
                        # Handle both single float and array returns
                        if isinstance(prob, (list, tuple)) or hasattr(prob, '__getitem__'):
                            if len(prob) > 0:
                                if hasattr(prob[0], '__len__') and len(prob[0]) > 1:
                                    prob = float(prob[0][1])  # Binary classification - positive class
                                else:
                                    prob = float(prob[0])
                            else:
                                prob = 0.5
                        else:
                            prob = float(prob)
                        
                        # Ensure we have a valid float
                        if prob is None or not isinstance(prob, (int, float)):
                            raise ValueError(f"Classifier returned invalid probability: {prob} (type: {type(prob)}). Expected float but got {type(prob).__name__}")
                        
                        # Additional validation
                        if not (0.0 <= prob <= 1.0):
                            raise ValueError(f"Classifier probability {prob} is out of valid range [0.0, 1.0]")
                            
                        token_scores.append(float(prob))
                    except Exception as e:
                        if verbose:
                            print(f"      ⚠️  Classifier error for token {token_idx}: {e}")
                        token_scores.append(0.5)  # Neutral score
                else:
                    token_scores.append(0.5)  # Neutral score if no classifier
    
    except Exception as e:
        if verbose:
            print(f"      ⚠️  Error during token scoring: {e}")
        # Fallback: assign neutral scores
        response_tokens = model.tokenizer(response, return_tensors="pt")['input_ids'][0]
        token_scores = [0.5] * len(response_tokens)
    
    # Classify overall response using specified aggregation method
    if token_scores:
        aggregated_score = aggregate_token_scores(token_scores, token_aggregation)
        classification = "HALLUCINATION" if aggregated_score > threshold else "TRUTHFUL"
    else:
        aggregated_score = 0.5
        classification = "UNKNOWN"
    
    return response, token_scores, classification


def generate_with_multi_layer_classification(model, prompt, layers, max_new_tokens, steering_methods, token_aggregation="average", threshold=0.6, verbose=False):
    """
    Generate text with token-level classification across multiple layers.
    
    Args:
        model: Model object
        prompt: Input prompt
        layers: List of layer indices for activation extraction
        max_new_tokens: Maximum tokens to generate
        steering_methods: Dict mapping layer indices to trained steering methods
        token_aggregation: How to aggregate token scores ("average", "final", "first", "max", "min")
        threshold: Classification threshold
        verbose: Whether to print debug info
        
    Returns:
        Tuple of (response_text, layer_results_dict)
        where layer_results_dict = {layer: {"token_scores": [...], "classification": "..."}}
    """
    import torch
    from .core import Layer
    from .core.activations import Activations, ActivationAggregationMethod
    from .core.parser import aggregate_token_scores
    
    # Generate response once (same for all layers)
    response, _ = model.generate(prompt, layers[0], max_new_tokens)
    
    if not response.strip():
        # Return empty results for all layers
        layer_results = {}
        for layer in layers:
            layer_results[layer] = {
                "token_scores": [],
                "classification": "UNKNOWN",
                "aggregated_score": 0.5
            }
        return response, layer_results
    
    # Tokenize the full prompt + response to get individual tokens
    full_text = f"{prompt}{response}"
    inputs = model.tokenizer(full_text, return_tensors="pt")
    tokens = model.tokenizer.convert_ids_to_tokens(inputs['input_ids'][0])
    
    # Get prompt length to identify response tokens
    prompt_inputs = model.tokenizer(prompt, return_tensors="pt")
    prompt_length = len(prompt_inputs['input_ids'][0])
    
    # Get model device and move inputs
    model_device = next(model.hf_model.parameters()).device
    inputs_on_device = {k: v.to(model_device) for k, v in inputs.items()}
    
    # Get hidden states for all layers at once
    with torch.no_grad():
        outputs = model.hf_model(**inputs_on_device, output_hidden_states=True)
    
    # Process each layer
    layer_results = {}
    
    for layer in layers:
        if verbose:
            print(f"    Processing layer {layer}...")
        
        # Extract activations for this layer
        token_scores = []
        layer_obj = Layer(index=layer, type="transformer")
        
        try:
            # Get hidden states for this layer
            if layer + 1 < len(outputs.hidden_states):
                hidden_states = outputs.hidden_states[layer + 1]
            else:
                hidden_states = outputs.hidden_states[-1]
            
            # Score each response token for this layer
            for token_idx in range(prompt_length, len(tokens)):
                if token_idx < hidden_states.shape[1]:
                    # Extract activation for this token
                    token_activation = hidden_states[0, token_idx, :].cpu()
                    
                    # Create Activations object
                    activation_obj = Activations(
                        tensor=token_activation.unsqueeze(0),
                        layer=layer_obj,
                        aggregation_method=ActivationAggregationMethod.LAST_TOKEN
                    )
                    
                    # Get feature vector for classifier
                    features = activation_obj.extract_features_for_classifier()
                    
                    # Get prediction probability from classifier for this layer
                    steering_method = steering_methods.get(layer)
                    if steering_method and hasattr(steering_method, 'classifier') and steering_method.classifier:
                        try:
                            prob = steering_method.classifier.predict_proba([features.numpy()])
                            # Handle both single float and array returns
                            if isinstance(prob, (list, tuple)) or hasattr(prob, '__getitem__'):
                                prob = float(prob[0]) if len(prob) > 0 else 0.5
                            else:
                                prob = float(prob)
                            token_scores.append(prob)
                        except Exception as e:
                            if verbose:
                                print(f"      ⚠️  Layer {layer} classifier error for token {token_idx}: {e}")
                            token_scores.append(0.5)
                    else:
                        token_scores.append(0.5)
        
        except Exception as e:
            if verbose:
                print(f"      ⚠️  Error during layer {layer} token scoring: {e}")
            # Fallback: assign neutral scores
            response_tokens = model.tokenizer(response, return_tensors="pt")['input_ids'][0]
            token_scores = [0.5] * len(response_tokens)
        
        # Classify overall response for this layer
        if token_scores:
            aggregated_score = aggregate_token_scores(token_scores, token_aggregation)
            classification = "HALLUCINATION" if aggregated_score > threshold else "TRUTHFUL"
        else:
            aggregated_score = 0.5
            classification = "UNKNOWN"
        
        layer_results[layer] = {
            "token_scores": token_scores,
            "classification": classification,
            "aggregated_score": aggregated_score
        }
        
        if verbose:
            print(f"      Layer {layer}: {classification} (score: {aggregated_score:.3f})")
    
    return response, layer_results


def generate_with_multi_layer_classification_and_handling(
    model, prompt, layers, max_new_tokens, steering_methods, 
    token_aggregation="average", threshold=0.6, verbose=False,
    detection_handler=None
):
    """
    Generate text with multi-layer classification and optional detection handling.
    
    Args:
        model: Model object
        prompt: Input prompt
        layers: List of layer indices for activation extraction
        max_new_tokens: Maximum tokens to generate
        steering_methods: Dict mapping layer indices to trained steering methods
        token_aggregation: How to aggregate token scores
        threshold: Classification threshold
        verbose: Whether to print debug info
        detection_handler: Optional DetectionHandler for handling detected issues
        
    Returns:
        Tuple of (final_response_text, layer_results_dict, was_handled)
    """
    from .core.detection_handling import DetectionHandler, DetectionAction
    from .core.parser import aggregate_token_scores
    
    # Generate initial response with multi-layer classification
    original_response, layer_results = generate_with_multi_layer_classification(
        model, prompt, layers, max_new_tokens, steering_methods, token_aggregation, threshold, verbose
    )
    
    # If no handler provided, return as-is
    if not detection_handler:
        return original_response, layer_results, False
    
    # Determine if content should be handled based on ANY layer detecting issues
    # Use the first layer for primary detection decision (could be made configurable)
    primary_layer = layers[0]
    is_problematic = layer_results[primary_layer]["classification"] == "HALLUCINATION"
    
    if not is_problematic:
        return original_response, layer_results, False
    
    # Content is problematic, apply handling
    if verbose:
        print(f"      🚨 Detected problematic content (layer {primary_layer}), applying {detection_handler.action.value} handling...")
    
    # Calculate confidence score for the detection using primary layer
    confidence_score = layer_results[primary_layer]["aggregated_score"]
    
    # Create regeneration function if needed
    regenerate_function = None
    if detection_handler.action == DetectionAction.REGENERATE_UNTIL_SAFE:
        def regenerate():
            # Generate a new response
            new_response, new_layer_results = generate_with_multi_layer_classification(
                model, prompt, layers, max_new_tokens, steering_methods, token_aggregation, threshold, verbose
            )
            # Only return if primary layer is not problematic
            if new_layer_results[primary_layer]["classification"] != "HALLUCINATION":
                return new_response
            else:
                # Still problematic, raise exception to trigger retry
                raise Exception(f"Generated response still classified as {new_layer_results[primary_layer]['classification']}")
        
        regenerate_function = regenerate
    
    # Handle the detection
    final_response = detection_handler.handle_detection(
        original_response=original_response,
        detection_type="hallucination",
        confidence_score=confidence_score,
        original_prompt=prompt,
        regenerate_function=regenerate_function
    )
    
    # If regeneration was successful, get new results
    if (detection_handler.action == DetectionAction.REGENERATE_UNTIL_SAFE 
        and final_response != original_response 
        and not final_response.startswith("I apologize")):
        
        # Re-classify the final response across all layers
        final_response_clean, final_layer_results = generate_with_multi_layer_classification(
            model, prompt, layers, max_new_tokens, steering_methods, token_aggregation, threshold, verbose=False
        )
        
        if final_response_clean.strip() == final_response.strip():
            return final_response, final_layer_results, True
    
    # For placeholder or pass-through, return original results but indicate handling occurred
    return final_response, layer_results, True


"""
LM-Evaluation-Harness task wrapper for task-agnostic architecture.
"""

from typing import Dict, Any, List, Optional
from ..task_interface import TaskInterface
from ..benchmark_extractors import BenchmarkExtractor, get_extractor


class LMEvalTask(TaskInterface):
    """Wrapper for lm-evaluation-harness tasks."""
    
    def __init__(self, task_name: str, description: str, categories: List[str]):
        self.task_name = task_name
        self._description = description
        self._categories = categories
        self._extractor = get_extractor(task_name)
    
    def load_data(self, limit: Optional[int] = None) -> List[Dict[str, Any]]:
        """Load data using lm-eval."""
        try:
            # Import here to avoid circular dependencies
            from ..model import Model
            
            # Create a temporary model instance to load task data
            # In a real implementation, this would be optimized
            model = Model("dummy")  # We don't actually need the model for data loading
            task_data = model.load_lm_eval_task(self.task_name, limit=limit)
            
            # Convert lm-eval task data to our format
            if hasattr(task_data, 'docs'):
                docs = task_data.docs
                if limit:
                    docs = docs[:limit]
                return docs
            else:
                return []
        except Exception as e:
            print(f"Warning: Could not load lm-eval task '{self.task_name}': {e}")
            return []
    
    def get_extractor(self) -> BenchmarkExtractor:
        """Get the benchmark extractor for this task."""
        return self._extractor
    
    def get_name(self) -> str:
        """Get the task name."""
        return self.task_name
    
    def get_description(self) -> str:
        """Get the task description."""
        return self._description
    
    def get_categories(self) -> List[str]:
        """Get the task categories."""
        return self._categories


class MBPPTask(LMEvalTask):
    """MBPP task implementation."""
    
    def __init__(self):
        super().__init__(
            task_name="mbpp",
            description="MBPP: Mostly Basic Python Problems coding benchmark",
            categories=["coding", "reasoning", "python"]
        )


class GSM8KTask(LMEvalTask):
    """GSM8K task implementation."""
    
    def __init__(self):
        super().__init__(
            task_name="gsm8k",
            description="GSM8K: Grade School Math 8K problems",
            categories=["mathematics", "reasoning", "arithmetic"]
        )


class TruthfulQATask(LMEvalTask):
    """TruthfulQA task implementation."""
    
    def __init__(self):
        super().__init__(
            task_name="truthfulqa_mc1",
            description="TruthfulQA: Truthfulness evaluation benchmark",
            categories=["hallucination", "general-knowledge", "reasoning"]
        )


class MMLUTask(LMEvalTask):
    """MMLU task implementation."""
    
    def __init__(self):
        super().__init__(
            task_name="mmlu",
            description="MMLU: Massive Multitask Language Understanding",
            categories=["general-knowledge", "science", "reasoning"]
        )


"""
Householder Pseudo-Rotation (HPR) steering method.

This implementation follows the reference HPR approach which uses:
1. A binary classifier to distinguish positive/negative activations
2. An angle prediction network to determine rotation amounts
3. Selective application of Householder transformations
"""

from typing import Dict, Any, Optional
import torch
import torch.nn as nn
import torch.nn.functional as F
import torch.optim as optim
import numpy as np

from .base import SteeringMethod
from ..contrastive_pairs import ContrastivePairSet


class BinaryClassifier(nn.Module):
    """Linear classifier to distinguish positive from negative activations."""
    
    def __init__(self, hidden_dim: int):
        super().__init__()
        self.linear = nn.Linear(hidden_dim, 1, bias=True)
        
    def forward(self, x: torch.Tensor) -> torch.Tensor:
        return self.linear(x)
    
    def get_hyperplane_normal(self) -> torch.Tensor:
        """Get the normal vector of the classification hyperplane."""
        return self.linear.weight.squeeze(0)


class AnglePredictor(nn.Module):
    """Neural network to predict rotation angles for activations."""
    
    def __init__(self, hidden_dim: int, hidden_size: int = 128):
        super().__init__()
        self.net = nn.Sequential(
            nn.Linear(hidden_dim, hidden_size),
            nn.ReLU(),
            nn.Linear(hidden_size, hidden_size),
            nn.ReLU(),
            nn.Linear(hidden_size, 1)
        )
        
    def forward(self, x: torch.Tensor) -> torch.Tensor:
        # Predict angles in radians
        return self.net(x)


class HPR(SteeringMethod):
    """
    Householder Pseudo-Rotation (HPR) steering method.
    
    Uses learned classifiers and angle predictors to selectively rotate
    activations in a norm-preserving manner.
    """
    
    def __init__(
        self, 
        device: Optional[str] = None,
        learning_rate: float = 1e-3,
        epochs: int = 100,
        batch_size: int = 32,
        hidden_size: int = 128,
        angle_loss_weight: float = 0.1
    ):
        super().__init__("HPR", device)
        self.classifier = None
        self.angle_predictor = None
        self.householder_matrix = None
        self.layer_index = None
        self.training_stats = {}
        
        # Training hyperparameters
        self.learning_rate = learning_rate
        self.epochs = epochs
        self.batch_size = batch_size
        self.hidden_size = hidden_size
        self.angle_loss_weight = angle_loss_weight
        
    def train(self, contrastive_pair_set: ContrastivePairSet, layer_index: int) -> Dict[str, Any]:
        """
        Train HPR components: classifier and angle predictor.
        
        Args:
            contrastive_pair_set: Set of contrastive pairs with activations
            layer_index: Layer index where steering will be applied
            
        Returns:
            Dictionary with training statistics
        """
        self.layer_index = layer_index
        
        # Get positive and negative activations
        pos_activations, neg_activations = contrastive_pair_set.get_activation_pairs()
        
        # Convert to tensors and create labels
        pos_tensors = torch.stack([act.to(self.device) for act in pos_activations])
        neg_tensors = torch.stack([act.to(self.device) for act in neg_activations])
        
        # Flatten if needed (we work with hidden states)
        if len(pos_tensors.shape) > 2:
            pos_tensors = pos_tensors.view(pos_tensors.shape[0], -1)
            neg_tensors = neg_tensors.view(neg_tensors.shape[0], -1)
        
        hidden_dim = pos_tensors.shape[1]
        
        # Create dataset
        all_activations = torch.cat([pos_tensors, neg_tensors], dim=0)
        labels = torch.cat([
            torch.ones(len(pos_tensors), 1),
            torch.zeros(len(neg_tensors), 1)
        ], dim=0).to(self.device)
        
        # Initialize models
        self.classifier = BinaryClassifier(hidden_dim).to(self.device)
        self.angle_predictor = AnglePredictor(hidden_dim, self.hidden_size).to(self.device)
        
        # Optimizers
        classifier_optimizer = optim.Adam(self.classifier.parameters(), lr=self.learning_rate)
        angle_optimizer = optim.Adam(self.angle_predictor.parameters(), lr=self.learning_rate)
        
        # Training loop
        dataset_size = all_activations.shape[0]
        losses = {'classifier': [], 'angle': [], 'total': []}
        
        for epoch in range(self.epochs):
            # Shuffle data
            perm = torch.randperm(dataset_size)
            all_activations = all_activations[perm]
            labels = labels[perm]
            
            epoch_losses = {'classifier': 0, 'angle': 0, 'total': 0}
            num_batches = 0
            
            for i in range(0, dataset_size, self.batch_size):
                batch_acts = all_activations[i:i + self.batch_size]
                batch_labels = labels[i:i + self.batch_size]
                
                # Classifier forward pass
                classifier_optimizer.zero_grad()
                logits = self.classifier(batch_acts)
                classifier_loss = F.binary_cross_entropy_with_logits(logits, batch_labels)
                
                # Angle predictor forward pass
                angle_optimizer.zero_grad()
                predicted_angles = self.angle_predictor(batch_acts)
                
                # Angle loss: encourage larger angles for negative examples
                # and smaller angles for positive examples
                target_angles = torch.where(
                    batch_labels > 0.5,
                    torch.zeros_like(predicted_angles),  # Positive: no rotation
                    torch.ones_like(predicted_angles) * np.pi / 4  # Negative: rotate by π/4
                )
                angle_loss = F.mse_loss(predicted_angles, target_angles)
                
                # Combined loss
                total_loss = classifier_loss + self.angle_loss_weight * angle_loss
                
                # Backward pass
                total_loss.backward()
                classifier_optimizer.step()
                angle_optimizer.step()
                
                # Record losses
                epoch_losses['classifier'] += classifier_loss.item()
                epoch_losses['angle'] += angle_loss.item()
                epoch_losses['total'] += total_loss.item()
                num_batches += 1
            
            # Average losses for epoch
            for key in epoch_losses:
                epoch_losses[key] /= num_batches
                losses[key].append(epoch_losses[key])
        
        # Compute Householder matrix from classifier hyperplane
        hyperplane_normal = self.classifier.get_hyperplane_normal()
        hyperplane_normal = F.normalize(hyperplane_normal, dim=0)
        
        # Householder matrix: H = I - 2 * v * v^T
        identity = torch.eye(hidden_dim, device=self.device)
        self.householder_matrix = identity - 2.0 * torch.outer(hyperplane_normal, hyperplane_normal)
        
        # Compute final accuracy
        with torch.no_grad():
            predictions = torch.sigmoid(self.classifier(all_activations)) > 0.5
            accuracy = (predictions == labels).float().mean().item()
        
        # Training statistics
        self.training_stats = {
            'method': 'HPR',
            'layer_index': layer_index,
            'hidden_dim': hidden_dim,
            'epochs': self.epochs,
            'final_accuracy': accuracy,
            'final_classifier_loss': losses['classifier'][-1],
            'final_angle_loss': losses['angle'][-1],
            'hyperplane_norm': torch.norm(hyperplane_normal).item(),
            'householder_matrix_norm': torch.norm(self.householder_matrix).item()
        }
        
        self.is_trained = True
        return self.training_stats
    
    def apply_steering(
        self, 
        activations: torch.Tensor, 
        strength: float = 1.0,
        apply_to_all: bool = False
    ) -> torch.Tensor:
        """
        Apply HPR steering using learned components.
        
        Args:
            activations: Input activations to steer
            strength: Steering strength multiplier
            apply_to_all: If True, apply to all activations (not just negative ones)
            
        Returns:
            Steered activations with preserved norms
        """
        if not self.is_trained:
            raise ValueError("HPR method must be trained before applying steering")
        
        original_shape = activations.shape
        device = activations.device
        
        # Flatten to 2D for processing
        if len(original_shape) == 3:  # [batch, seq, hidden]
            batch_size, seq_len, hidden_dim = original_shape
            # Apply to second-to-last token position (reference behavior)
            if seq_len > 1:
                # Extract the token we want to steer
                target_acts = activations[:, -2, :]  # [batch, hidden]
            else:
                target_acts = activations[:, -1, :]
        else:  # [batch, hidden]
            target_acts = activations
            batch_size = target_acts.shape[0]
        
        # Move models to correct device
        self.classifier = self.classifier.to(device)
        self.angle_predictor = self.angle_predictor.to(device)
        householder_matrix = self.householder_matrix.to(device)
        
        with torch.no_grad():
            # Classify activations
            logits = self.classifier(target_acts)
            is_negative = torch.sigmoid(logits).squeeze(1) < 0.5  # Shape: [batch]
            
            if not apply_to_all:
                # Only process negative activations
                mask = is_negative.float()
            else:
                # Process all activations
                mask = torch.ones_like(is_negative, dtype=torch.float)
            
            # Predict rotation angles
            angles = self.angle_predictor(target_acts).squeeze(1)  # Shape: [batch]
            angles = angles * strength * mask  # Scale by strength and mask
            
            # Apply rotation using Householder matrix
            # For each activation, we rotate by the predicted angle
            steered_acts = target_acts.clone()
            
            for i in range(batch_size):
                if mask[i] > 0:  # Only rotate if masked
                    angle = angles[i]
                    act = target_acts[i]
                    
                    # Householder reflection followed by rotation
                    # First reflect across hyperplane
                    reflected = torch.matmul(householder_matrix, act)
                    
                    # Then rotate by angle (simplified rotation in reflection space)
                    # This is a simplified version - full implementation would use
                    # proper rotation matrices in the reflection space
                    cos_angle = torch.cos(angle)
                    sin_angle = torch.sin(angle)
                    
                    # Interpolate between original and reflected based on angle
                    steered_acts[i] = cos_angle * act + sin_angle * reflected
        
        # Reconstruct full tensor
        if len(original_shape) == 3:
            result = activations.clone()
            if seq_len > 1:
                result[:, -2, :] = steered_acts
            else:
                result[:, -1, :] = steered_acts
            return result
        else:
            return steered_acts
    
    def get_steering_vector(self) -> torch.Tensor:
        """Return the hyperplane normal (for compatibility)."""
        if not self.is_trained:
            raise ValueError("HPR method must be trained before getting steering vector")
        return self.classifier.get_hyperplane_normal()
    
    def save_steering_vector(self, path: str) -> bool:
        """Save HPR steering components."""
        if not self.is_trained:
            return False
        try:
            torch.save({
                'classifier_state': self.classifier.state_dict(),
                'angle_predictor_state': self.angle_predictor.state_dict(),
                'householder_matrix': self.householder_matrix,
                'layer_index': self.layer_index,
                'training_stats': self.training_stats,
                'hidden_dim': self.householder_matrix.shape[0],
                'hidden_size': self.hidden_size,
                'method': 'HPR'
            }, path)
            return True
        except Exception:
            return False
    
    def load_steering_vector(self, path: str) -> bool:
        """Load HPR steering components."""
        try:
            data = torch.load(path, map_location=self.device, weights_only=False)
            if data.get('method') != 'HPR':
                return False
            
            # Recreate models
            hidden_dim = data['hidden_dim']
            self.hidden_size = data.get('hidden_size', 128)
            self.classifier = BinaryClassifier(hidden_dim).to(self.device)
            self.angle_predictor = AnglePredictor(hidden_dim, self.hidden_size).to(self.device)
            
            # Load states
            self.classifier.load_state_dict(data['classifier_state'])
            self.angle_predictor.load_state_dict(data['angle_predictor_state'])
            self.householder_matrix = data['householder_matrix']
            self.layer_index = data.get('layer_index')
            self.training_stats = data.get('training_stats', {})
            
            self.is_trained = True
            return True
        except Exception:
            return False


"""
Wisent Guard: Activation-based safety guardrails for language models.
Clean implementation using enhanced core primitives.
"""

from .guard import WisentGuard
from .inference import SafeInference
from .vectors import ContrastiveVectors

# Import from core primitives
from .core import (
    Model,
    PromptFormat,
    TokenScore,
    ModelParameterOptimizer,
    ActivationHooks,
    Activations,
    ActivationAggregationMethod,
    Layer,
    ContrastivePairSet,
    Classifier,
    ActivationClassifier,
    SteeringMethod
)

__version__ = "0.4.3"

__all__ = [
    # Main classes
    "WisentGuard",
    "SafeInference", 
    "ContrastiveVectors",
    
    # Core primitives
    "Model",
    "PromptFormat",
    "TokenScore", 
    "ModelParameterOptimizer",
    "ActivationHooks",
    "Activations",
    "ActivationAggregationMethod",
    "Layer",
    "ContrastivePairSet",
    "Classifier",
    "ActivationClassifier",
    "SteeringMethod"
] 


"""
Tests for the ContrastiveVectors class
"""

import os
import torch
import pytest
import tempfile
import shutil
from wisent_guard.vectors import ContrastiveVectors


@pytest.mark.slow
@pytest.mark.unit
class TestContrastiveVectors:
    """Test suite for ContrastiveVectors class."""
    
    @pytest.fixture(autouse=True)
    def setup_vectors(self):
        """Set up test fixtures."""
        self.test_dir = tempfile.mkdtemp()
        
        yield
        
        # Tear down test fixtures
        shutil.rmtree(self.test_dir)
    
    def test_vectors_initialization(self):
        """Test that ContrastiveVectors initializes correctly."""
        # Use a small model for testing
        vectors = ContrastiveVectors(
            model_name="distilbert/distilgpt2",
            layers=[0, 1],
            save_dir=self.test_dir
        )
        
        # Check that initialization worked
        assert vectors.model is not None
        assert len(vectors.layers) == 2
        assert vectors.save_dir.exists()
        
        # Check that layers are correctly set
        layer_indices = [layer.index for layer in vectors.layers]
        assert layer_indices == [0, 1]
    
    def test_create_pair_set(self):
        """Test creating a contrastive pair set."""
        vectors = ContrastiveVectors(
            model_name="distilbert/distilgpt2",
            layers=[0],
            save_dir=self.test_dir
        )
        
        # Create a simple pair set
        harmful_texts = ["This is harmful text", "Another harmful example"]
        harmless_texts = ["This is harmless text", "Another harmless example"]
        
        pair_set = vectors.create_pair_set(
            name="test_pairs",
            harmful_texts=harmful_texts,
            harmless_texts=harmless_texts
        )
        
        # Check that pair set was created
        assert pair_set is not None
        assert "test_pairs" in vectors.pair_sets
        assert len(pair_set.phrase_pairs) == 2
    
    def test_compute_vectors_basic(self):
        """Test basic vector computation."""
        vectors = ContrastiveVectors(
            model_name="distilbert/distilgpt2",
            layers=[0],
            save_dir=self.test_dir
        )
        
        # Create a simple pair set
        harmful_texts = ["Bad content"]
        harmless_texts = ["Good content"]
        
        vectors.create_pair_set(
            name="test_pairs",
            harmful_texts=harmful_texts,
            harmless_texts=harmless_texts
        )
        
        # Compute vectors
        computed_vectors = vectors.compute_vectors("test_pairs", layer_idx=0)
        
        # Check that vectors were computed
        assert isinstance(computed_vectors, dict)
        assert 0 in computed_vectors  # Layer 0 should have a vector
        
        # Check that the vector is stored
        assert "test_pairs" in vectors.vectors
        assert 0 in vectors.vectors["test_pairs"]
    
    def test_get_similarity_basic(self):
        """Test basic similarity computation."""
        vectors = ContrastiveVectors(
            model_name="distilbert/distilgpt2",
            layers=[0],
            save_dir=self.test_dir
        )
        
        # Create and compute vectors
        harmful_texts = ["Bad content"]
        harmless_texts = ["Good content"]
        
        vectors.create_pair_set(
            name="test_pairs",
            harmful_texts=harmful_texts,
            harmless_texts=harmless_texts
        )
        
        vectors.compute_vectors("test_pairs", layer_idx=0)
        
        # Test similarity
        similarity = vectors.get_similarity(
            text="Some test text",
            category="test_pairs",
            layer_idx=0
        )
        
        # Check that similarity is a number
        assert isinstance(similarity, (int, float))
    
    def test_vectors_storage_structure(self):
        """Test that vectors are stored in correct structure."""
        vectors = ContrastiveVectors(
            model_name="distilbert/distilgpt2",
            layers=[0, 1],
            save_dir=self.test_dir
        )
        
        # Create pair set
        vectors.create_pair_set(
            name="test_pairs",
            harmful_texts=["Bad"],
            harmless_texts=["Good"]
        )
        
        # Compute for all layers
        vectors.compute_vectors("test_pairs")
        
        # Check storage structure
        assert "test_pairs" in vectors.vectors
        assert isinstance(vectors.vectors["test_pairs"], dict)
        
        # Should have computed vectors for both layers
        available_layers = list(vectors.vectors["test_pairs"].keys())
        assert len(available_layers) > 0  # At least one layer should work
    
    def test_layer_configuration(self):
        """Test different layer configurations."""
        # Test with single layer
        vectors_single = ContrastiveVectors(
            model_name="distilbert/distilgpt2",
            layers=[0],
            save_dir=self.test_dir
        )
        assert len(vectors_single.layers) == 1
        
        # Test with multiple layers
        vectors_multi = ContrastiveVectors(
            model_name="distilbert/distilgpt2",
            layers=[0, 1, 2],
            save_dir=self.test_dir
        )
        assert len(vectors_multi.layers) == 3
        
        # Check layer indices
        layer_indices = [layer.index for layer in vectors_multi.layers]
        assert layer_indices == [0, 1, 2]
    
    def test_save_directory_creation(self):
        """Test that save directory is created correctly."""
        # Use a nested path
        nested_path = os.path.join(self.test_dir, "nested", "vectors")
        
        vectors = ContrastiveVectors(
            model_name="distilbert/distilgpt2",
            layers=[0],
            save_dir=nested_path
        )
        
        # Check that directory was created
        assert os.path.exists(nested_path)
        assert vectors.save_dir.exists()
        assert vectors.save_dir.is_dir()


class Response:
    def __init__(self, text, activations=None, label=None):
        self.text = text
        self.activations = activations
        self.label = label

class PositiveResponse(Response):
    pass

class NegativeResponse(Response):
    pass 


#!/usr/bin/env python3

import sys
import torch
from pathlib import Path
from transformers import AutoTokenizer, AutoModelForCausalLM

project_root = Path.cwd().parent
sys.path.insert(0, str(project_root))

from wisent_guard.core.steering_methods.dac import DAC
from wisent_guard.core.steering_methods.caa import CAA
from wisent_guard.core.steering_methods.hpr import HPR
from wisent_guard.core.steering_methods.bipo import BiPO
from wisent_guard.core.steering_methods.k_steering import KSteering
from wisent_guard.core.contrastive_pairs.contrastive_pair_set import ContrastivePairSet
from wisent_guard.core.contrastive_pairs.contrastive_pair import ContrastivePair
from wisent_guard.core.response import PositiveResponse, NegativeResponse

# Model configuration
MODEL_NAME = "meta-llama/Llama-3.1-8B-Instruct"
LAYER_INDEX = 15
STEERING_STRENGTH = 2.0
MAX_LENGTH = 50
NUM_PAIRS = 5  # Reduced for faster testing

print(f"Testing {MODEL_NAME} on layer {LAYER_INDEX}")

# Get optimal device
device = torch.device("mps" if torch.backends.mps.is_available() else 
                     "cuda" if torch.cuda.is_available() else "cpu")
print(f"Using device: {device}")

try:
    # Load model and tokenizer
    print("Loading model and tokenizer...")
    tokenizer = AutoTokenizer.from_pretrained(MODEL_NAME)
    model = AutoModelForCausalLM.from_pretrained(MODEL_NAME, torch_dtype=torch.float16 if device.type == 'cuda' else torch.float32).to(device)
    tokenizer.pad_token = tokenizer.eos_token
    print("✓ Model loaded successfully")

    # Test prompts
    TEST_PROMPTS = ["How can I help someone?", "What should I do?"]

    # Create contrastive pairs for harmfulness
    helpful_examples = [
        "I want to help people and be kind.",
        "Let me assist you with that safely.",
        "I can provide helpful information.",
        "Here's a constructive approach.",
        "I'd be happy to help responsibly."
    ]

    harmful_examples = [
        "I want to hurt people and cause damage.",
        "Let me help you do something dangerous.",
        "I can provide harmful information.",
        "Here's a destructive approach.",
        "I'd be happy to help with bad things."
    ]

    # Extract activations for contrastive pairs
    def extract_activations(text, layer_idx):
        inputs = tokenizer(text, return_tensors="pt", truncation=True, max_length=50).to(device)
        
        # Hook to capture activations
        activations = []
        def hook(module, input, output):
            activations.append(output[0][:, -1, :].clone())  # Last token
        
        # Register hook
        handle = model.model.layers[layer_idx].register_forward_hook(hook)
        
        with torch.no_grad():
            model(**inputs)
        
        handle.remove()
        return activations[0].squeeze(0)

    print("Extracting activations...")
    # Create contrastive pair set
    pair_set = ContrastivePairSet(name="harmfulness")

    for i in range(NUM_PAIRS):
        helpful_text = helpful_examples[i % len(helpful_examples)]
        harmful_text = harmful_examples[i % len(harmful_examples)]
        
        # Extract real activations
        helpful_activation = extract_activations(helpful_text, LAYER_INDEX)
        harmful_activation = extract_activations(harmful_text, LAYER_INDEX)
        
        # Create responses
        pos_resp = PositiveResponse(text=helpful_text)
        pos_resp.activations = helpful_activation
        
        neg_resp = NegativeResponse(text=harmful_text)
        neg_resp.activations = harmful_activation
        
        # Create pair
        pair = ContrastivePair(
            prompt=f"Respond helpfully: {helpful_text[:20]}...",
            positive_response=pos_resp,
            negative_response=neg_resp
        )
        pair_set.pairs.append(pair)

    print(f"✓ Created {len(pair_set.pairs)} contrastive pairs")

    # Train steering methods
    print("Training steering methods...")
    
    print("  Training DAC...")
    dac = DAC(device=device)
    dac.set_model_reference(model)
    dac.train(pair_set, LAYER_INDEX)
    
    print("  Training CAA...")
    caa = CAA(device=device)
    caa.train(pair_set, LAYER_INDEX)
    
    print("  Training HPR...")
    hpr = HPR(device=device, epochs=10)  # Reduced epochs for speed
    hpr.train(pair_set, LAYER_INDEX)
    
    print("  Training BiPO...")
    bipo = BiPO(device=device, num_epochs=10, batch_size=2)  # Reduced for speed
    bipo.train(pair_set, LAYER_INDEX)
    
    print("  Training K-Steering...")
    k_steering = KSteering(device=device, num_labels=1, classifier_epochs=50)  # Reduced epochs
    k_steering.train(pair_set, LAYER_INDEX)
    
    print("✓ All steering methods trained")

    # Generate responses with steering
    class SteeringModelWrapper:
        def __init__(self, model, tokenizer, steering_method, layer_idx, strength):
            self.model = model
            self.tokenizer = tokenizer
            self.steering_method = steering_method
            self.layer_idx = layer_idx
            self.strength = strength
            self.hooks = []
            
        def add_steering_hook(self, direction="positive"):
            def steering_hook(module, input, output):
                hidden_states = output[0]
                # Apply steering to last token
                last_token = hidden_states[:, -1:, :]
                if hasattr(self.steering_method, 'apply_steering'):
                    if direction == "negative" and hasattr(self.steering_method, 'get_bidirectional_vectors'):
                        steered = self.steering_method.apply_steering(last_token, self.strength, direction="negative")
                    else:
                        steered = self.steering_method.apply_steering(last_token, self.strength)
                    hidden_states[:, -1:, :] = steered
                return (hidden_states,) + output[1:]
            
            handle = self.model.model.layers[self.layer_idx].register_forward_hook(steering_hook)
            self.hooks.append(handle)
            
        def remove_hooks(self):
            for handle in self.hooks:
                handle.remove()
            self.hooks = []
            
        def generate(self, prompt, direction="positive"):
            self.add_steering_hook(direction)
            
            inputs = self.tokenizer(prompt, return_tensors="pt", truncation=True, max_length=30).to(device)
            
            with torch.no_grad():
                outputs = self.model.generate(
                    **inputs,
                    max_length=inputs['input_ids'].shape[1] + 20,  # Shorter for testing
                    do_sample=True,
                    temperature=0.7,
                    pad_token_id=self.tokenizer.eos_token_id
                )
            
            self.remove_hooks()
            
            response = self.tokenizer.decode(outputs[0], skip_special_tokens=True)
            return response[len(prompt):].strip()

    def generate_unsteered(prompt):
        inputs = tokenizer(prompt, return_tensors="pt", truncation=True, max_length=30).to(device)
        
        with torch.no_grad():
            outputs = model.generate(
                **inputs,
                max_length=inputs['input_ids'].shape[1] + 20,  # Shorter for testing
                do_sample=True,
                temperature=0.7,
                pad_token_id=tokenizer.eos_token_id
            )
        
        response = tokenizer.decode(outputs[0], skip_special_tokens=True)
        return response[len(prompt):].strip()

    print("\n" + "="*60)
    print("STEERING METHODS COMPARISON RESULTS")
    print("="*60)

    # Generate and display responses
    for prompt in TEST_PROMPTS:
        print(f"\nPrompt: {prompt}")
        print("-" * 40)
        
        # Unsteered
        unsteered = generate_unsteered(prompt)
        print(f"Unsteered: {unsteered}")
        
        # DAC
        dac_wrapper = SteeringModelWrapper(model, tokenizer, dac, LAYER_INDEX, STEERING_STRENGTH)
        dac_response = dac_wrapper.generate(prompt)
        print(f"DAC: {dac_response}")
        
        # CAA
        caa_wrapper = SteeringModelWrapper(model, tokenizer, caa, LAYER_INDEX, STEERING_STRENGTH)
        caa_response = caa_wrapper.generate(prompt)
        print(f"CAA: {caa_response}")
        
        # HPR
        hpr_wrapper = SteeringModelWrapper(model, tokenizer, hpr, LAYER_INDEX, STEERING_STRENGTH)
        hpr_response = hpr_wrapper.generate(prompt)
        print(f"HPR: {hpr_response}")
        
        # BiPO Positive
        bipo_wrapper = SteeringModelWrapper(model, tokenizer, bipo, LAYER_INDEX, STEERING_STRENGTH)
        bipo_pos_response = bipo_wrapper.generate(prompt, "positive")
        print(f"BiPO Positive: {bipo_pos_response}")
        
        # BiPO Negative
        bipo_neg_response = bipo_wrapper.generate(prompt, "negative")
        print(f"BiPO Negative: {bipo_neg_response}")
        
        # K-Steering
        k_wrapper = SteeringModelWrapper(model, tokenizer, k_steering, LAYER_INDEX, STEERING_STRENGTH)
        k_response = k_wrapper.generate(prompt)
        print(f"K-Steering: {k_response}")

    print("\n" + "="*60)
    print("✓ All steering methods tested successfully!")

except Exception as e:
    print(f"❌ Error: {e}")
    import traceback
    traceback.print_exc()


"""
Logging utilities for Wisent Guard
"""

import logging
import os
import sys
from typing import Optional

# Define log levels
LOG_LEVELS = {
    "debug": logging.DEBUG,
    "info": logging.INFO,
    "warning": logging.WARNING,
    "error": logging.ERROR,
    "critical": logging.CRITICAL
}

# ANSI color codes for terminal output
COLORS = {
    "reset": "\033[0m",
    "bold": "\033[1m",
    "red": "\033[31m",
    "green": "\033[32m",
    "yellow": "\033[33m",
    "blue": "\033[34m",
    "magenta": "\033[35m",
    "cyan": "\033[36m",
}

# Global logger instance
_logger = None

def get_logger(name: str = "wisent_guard", level: str = "info", 
               log_file: Optional[str] = None, use_colors: bool = True) -> logging.Logger:
    """
    Get or create a logger with the specified configuration.
    
    Args:
        name: Logger name
        level: Log level ('debug', 'info', 'warning', 'error', 'critical')
        log_file: Optional file path to write logs to
        use_colors: Whether to use colored terminal output
        
    Returns:
        Configured logger instance
    """
    global _logger
    
    if _logger is not None:
        return _logger
    
    # Create logger
    logger = logging.getLogger(name)
    
    # Set level from string
    log_level = LOG_LEVELS.get(level.lower(), logging.INFO)
    logger.setLevel(log_level)
    
    # Create handlers
    handlers = []
    
    # Console handler
    console_handler = logging.StreamHandler(sys.stdout)
    console_handler.setLevel(log_level)
    
    # Format with colors if enabled
    if use_colors:
        console_formatter = logging.Formatter(
            f"{COLORS['bold']}{COLORS['blue']}[%(asctime)s]{COLORS['reset']} "
            f"{COLORS['bold']}%(levelname)s:{COLORS['reset']} "
            f"{COLORS['cyan']}[%(name)s]{COLORS['reset']} %(message)s",
            datefmt="%H:%M:%S"
        )
    else:
        console_formatter = logging.Formatter(
            "[%(asctime)s] %(levelname)s: [%(name)s] %(message)s",
            datefmt="%H:%M:%S"
        )
    
    console_handler.setFormatter(console_formatter)
    handlers.append(console_handler)
    
    # File handler if log_file is specified
    if log_file:
        log_dir = os.path.dirname(log_file)
        if log_dir and not os.path.exists(log_dir):
            os.makedirs(log_dir)
            
        file_handler = logging.FileHandler(log_file)
        file_handler.setLevel(log_level)
        file_formatter = logging.Formatter(
            "[%(asctime)s] %(levelname)s: [%(name)s] %(message)s",
            datefmt="%Y-%m-%d %H:%M:%S"
        )
        file_handler.setFormatter(file_formatter)
        handlers.append(file_handler)
    
    # Add handlers to logger
    for handler in handlers:
        logger.addHandler(handler)
    
    # Store logger globally
    _logger = logger
    return logger

def set_log_level(level: str) -> None:
    """
    Set the log level for the existing logger.
    
    Args:
        level: Log level ('debug', 'info', 'warning', 'error', 'critical')
    """
    global _logger
    if _logger is not None:
        log_level = LOG_LEVELS.get(level.lower(), logging.INFO)
        _logger.setLevel(log_level)
        for handler in _logger.handlers:
            handler.setLevel(log_level) 


from transformers import AutoModelForCausalLM, AutoTokenizer
import torch
import torch.nn.functional as F
import re
from typing import Optional, Union, List, Dict, Any, Tuple
from enum import Enum
from .contrastive_pairs import ContrastivePairSet
from .user_model_config import user_model_configs

class PromptFormat(Enum):
    LEGACY = "legacy"
    LLAMA31 = "llama31"
    MISTRAL = "mistral"

class TokenScore:
    """Stores information about a token and its similarity to harmful content."""
    
    def __init__(self, 
                 token_id: Optional[int] = None,
                 token_text: str = "",
                 position: int = 0,
                 similarity: float = 0.0,
                 is_harmful: bool = False,
                 category: Optional[str] = None,
                 activations: Optional[Dict[int, torch.Tensor]] = None):
        self.token_id = token_id
        self.token_text = token_text
        self.position = position
        self.similarity = similarity
        self.is_harmful = is_harmful
        self.category = category
        self.activations = activations or {}
    
    def to_dict(self) -> Dict[str, Any]:
        """Convert to dictionary representation."""
        return {
            "token_id": self.token_id,
            "token_text": self.token_text,
            "position": self.position,
            "similarity": float(self.similarity),
            "is_harmful": self.is_harmful,
            "category": self.category,
        }

class Model:
    def __init__(self, name: str, layers: Optional[List[int]] = None, device: Optional[str] = None, hf_model: Optional[AutoModelForCausalLM] = None):
        """
        Initialize Model with either a model name to load or an existing HuggingFace model.
        
        Args:
            name: Model name for HuggingFace or identifier
            layers: List of layer indices to use
            device: Device to run on
            hf_model: Optional existing HuggingFace model to wrap
        """
        self.name = name
        self.layers = layers if layers is not None else []
        self.device = device or ("cuda" if torch.cuda.is_available() else "mps" if hasattr(torch.backends, 'mps') and torch.backends.mps.is_available() else "cpu")
        
        # Prompt formatting settings - will be set based on model type
        self.user_token = None
        self.assistant_token = None
        self.format_type = None  # Will be auto-detected
        
        if hf_model is not None:
            # Use provided model
            self.hf_model = hf_model
            self.model = hf_model  # Keep backward compatibility
            
            # Auto-detect format type first
            self.format_type = self._detect_format()
            
            # Try to load tokenizer from the same name
            try:
                self.tokenizer = AutoTokenizer.from_pretrained(name)
                if self.tokenizer.pad_token is None:
                    self.tokenizer.pad_token = self.tokenizer.eos_token
            except:
                self.tokenizer = None
                
            # Initialize tokens
            self._initialize_tokens()
        else:
            # Check if model is supported before trying to load
            self.hf_model = None
            self.model = None
            self.tokenizer = None
            
            # Auto-detect format type first (this will raise error if unsupported)
            self.format_type = self._detect_format()
            
            # Only load model after we know it's supported
            self._load_model_and_tokenizer()
            
            # Initialize tokens after loading
            self._initialize_tokens()

    def _load_model_and_tokenizer(self):
        """Load model and tokenizer from HuggingFace."""
        if self.device == "mps":
            torch_dtype = torch.float32
        else:
            torch_dtype = torch.float16 if self.device != "cpu" else torch.float32
            
        self.tokenizer = AutoTokenizer.from_pretrained(self.name)
        if self.tokenizer.pad_token is None:
            self.tokenizer.pad_token = self.tokenizer.eos_token
            
        self.hf_model = AutoModelForCausalLM.from_pretrained(
            self.name,
            torch_dtype=torch_dtype,
            device_map=self.device,
            output_hidden_states=True
        )
        self.hf_model.config.output_hidden_states = True
        self.hf_model.eval()
        
        # Keep backward compatibility
        self.model = self.hf_model

    def _detect_format(self) -> PromptFormat:
        """
        Detect the appropriate format to use based on the model.
        
        Returns:
            PromptFormat enum indicating the format
            
        Raises:
            ValueError: If model is not recognized and no user config exists
        """
        model_name = self.name.lower()
        
        # Check for specific model types
        if re.search(r"llama-?3", model_name, re.IGNORECASE):
            return PromptFormat.LLAMA31
        elif "mistral" in model_name:
            return PromptFormat.MISTRAL
        elif "gpt2" in model_name or "distilgpt2" in model_name:
            return PromptFormat.LEGACY
        else:
            # Check if we have a user-defined config
            if user_model_configs.has_config(self.name):
                # User has configured this model
                return PromptFormat.LEGACY  # We'll handle tokens separately
            else:
                # Unknown model with no config - prompt user to configure it
                print(f"\n⚠️  Model '{self.name}' is not recognized.")
                response = input("Would you like to configure it now? (y/n): ").strip().lower()
                
                if response == 'y':
                    # Configure the model interactively
                    config = user_model_configs.prompt_and_save_config(self.name)
                    return PromptFormat.LEGACY  # User-configured models use legacy format
                else:
                    raise ValueError(
                        f"Model '{self.name}' requires configuration.\n"
                        f"Run 'wisent-guard configure-model {self.name}' to set it up."
                    )

    def _initialize_tokens(self):
        """Initialize user and assistant tokens based on format type or user config."""
        if self.format_type == PromptFormat.LLAMA31:
            # Llama uses header-based format, not simple tokens
            self.user_token = "user"  # Used in header
            self.assistant_token = "assistant"
        elif self.format_type == PromptFormat.MISTRAL:
            # Mistral uses instruction format
            self.user_token = "[INST]"
            self.assistant_token = "[/INST]"
        elif self.format_type == PromptFormat.LEGACY:
            # Check for user-defined tokens
            if user_model_configs.has_config(self.name):
                tokens = user_model_configs.get_prompt_tokens(self.name)
                if tokens:
                    self.user_token = tokens.get("user_token", "<|user|>")
                    self.assistant_token = tokens.get("assistant_token", "<|assistant|>")
                else:
                    # Default legacy tokens
                    self.user_token = "<|user|>"
                    self.assistant_token = "<|assistant|>"
            else:
                # Default legacy tokens for known models like GPT2
                self.user_token = "<|user|>"
                self.assistant_token = "<|assistant|>"
    
    def set_prompt_tokens(self, user_token: str, assistant_token: str):
        """Set custom user and assistant tokens for legacy format."""
        self.user_token = user_token
        self.assistant_token = assistant_token

    def format_prompt(self, prompt: str, response: str = None) -> str:
        """
        Format a prompt using the appropriate format (Llama 3.1, Mistral, or legacy).
        
        Args:
            prompt: Input prompt text
            response: Optional response text to include
            
        Returns:
            Formatted prompt string
        """
        if self.format_type == PromptFormat.LLAMA31:
            # Llama 3.1 special tokens
            BEGIN_TEXT = "<|begin_of_text|>"
            START_HEADER = "<|start_header_id|>"
            END_HEADER = "<|end_header_id|>"
            EOT = "<|eot_id|>"
            
            if response is not None:
                return f"{BEGIN_TEXT}{START_HEADER}user{END_HEADER}\n{prompt}{EOT}{START_HEADER}assistant{END_HEADER}\n{response}"
            else:
                return f"{BEGIN_TEXT}{START_HEADER}user{END_HEADER}\n{prompt}{EOT}{START_HEADER}assistant{END_HEADER}\n"
        elif self.format_type == PromptFormat.MISTRAL:
            # Mistral format tokens
            INST_START = "[INST]"
            INST_END = "[/INST]"
            
            if response is not None:
                return f"{INST_START} {prompt} {INST_END} {response}"
            else:
                return f"{INST_START} {prompt} {INST_END}"
        else:
            # Legacy format
            if response is not None:
                return f"{self.user_token}\n{prompt}\n{self.assistant_token}\n{response}"
            else:
                return f"{self.user_token}\n{prompt}\n{self.assistant_token}\n"

    def format_multiple_choice(self, question: str, choice_a: str, choice_b: str, answer: str) -> str:
        """Format a multiple-choice question with the given answer."""
        instruction = f"{question}\nA. {choice_a}\nB. {choice_b}"
        return self.format_prompt(instruction, answer)

    def convert_to_multiple_choice(self, harmful_phrase: str, harmless_phrase: str) -> Tuple[str, str]:
        """
        Convert a harmful/harmless phrase pair to multiple-choice format.
        
        Args:
            harmful_phrase: The harmful phrase
            harmless_phrase: The harmless phrase
            
        Returns:
            Tuple of (harmful_mc, harmless_mc) in multiple-choice format
        """
        # Create question from the phrases
        question = "Which of the following statements is better?"
        
        # Format as multiple-choice with A as harmless and B as harmful
        harmful_mc = self.format_multiple_choice(
            question=question,
            choice_a=harmless_phrase,
            choice_b=harmful_phrase,
            answer="B"  # Harmful choice (B) is correct for harmful_mc
        )
        
        harmless_mc = self.format_multiple_choice(
            question=question,
            choice_a=harmless_phrase,
            choice_b=harmful_phrase,
            answer="A"  # Harmless choice (A) is correct for harmless_mc
        )
        
        return harmful_mc, harmless_mc

    def get_num_layers(self) -> int:
        """Get the number of layers in the model."""
        if self.hf_model is None:
            return 0
        
        # Try different model architectures
        if hasattr(self.hf_model, 'model') and hasattr(self.hf_model.model, 'layers'):
            return len(self.hf_model.model.layers)
        elif hasattr(self.hf_model, 'transformer') and hasattr(self.hf_model.transformer, 'h'):
            return len(self.hf_model.transformer.h)
        elif hasattr(self.hf_model.config, 'num_hidden_layers'):
            return self.hf_model.config.num_hidden_layers
        else:
            return 12  # Default fallback

    def generate_monitored(
        self,
        prompt: str,
        monitor_callback: Optional[callable] = None,
        max_new_tokens: int = 50,
        do_sample: bool = True,
        temperature: float = 0.7,
        top_p: float = 0.9,
        **kwargs
    ) -> Dict[str, Any]:
        """
        Generate text token-by-token with monitoring capabilities.
        
        Args:
            prompt: Input prompt string
            monitor_callback: Optional callback function for monitoring each token
            max_new_tokens: Maximum number of new tokens to generate
            do_sample: Whether to use sampling for generation
            temperature: Sampling temperature
            top_p: Top-p sampling parameter
            **kwargs: Additional generation parameters
            
        Returns:
            Dictionary containing generated text and token-level analysis
        """
        if self.hf_model is None:
            raise ValueError("No model loaded")
        
        # Format the prompt
        formatted_prompt = self.format_prompt(prompt)
        
        # Tokenize the prompt
        prompt_inputs = self.tokenizer(formatted_prompt, return_tensors="pt").to(self.device)
        prompt_ids = prompt_inputs.input_ids
        prompt_length = prompt_ids.shape[1]
        
        # Initialize result containers
        generated_tokens = []
        token_scores = []
        
        # Current input for generation
        current_input = prompt_ids.clone()
        
        # Generate tokens one by one
        for token_idx in range(max_new_tokens):
            # Generate next token
            with torch.no_grad():
                outputs = self.hf_model(
                    current_input,
                    output_hidden_states=True,
                    return_dict=True
                )
                
                # Get logits for the last token
                next_token_logits = outputs.logits[0, -1, :]
                
                # Apply temperature and top-p sampling
                if do_sample:
                    # Apply temperature
                    if temperature != 1.0:
                        next_token_logits = next_token_logits / temperature
                    
                    # Apply top-p sampling
                    if top_p < 1.0:
                        sorted_logits, sorted_indices = torch.sort(next_token_logits, descending=True)
                        cumulative_probs = torch.cumsum(F.softmax(sorted_logits, dim=-1), dim=-1)
                        
                        # Remove tokens with cumulative probability above the threshold
                        sorted_indices_to_remove = cumulative_probs > top_p
                        sorted_indices_to_remove[1:] = sorted_indices_to_remove[:-1].clone()
                        sorted_indices_to_remove[0] = 0
                        
                        indices_to_remove = sorted_indices[sorted_indices_to_remove]
                        next_token_logits[indices_to_remove] = float('-inf')
                    
                    # Sample from the distribution
                    probs = F.softmax(next_token_logits, dim=-1)
                    next_token = torch.multinomial(probs, num_samples=1)
                else:
                    # Greedy decoding
                    next_token = torch.argmax(next_token_logits, dim=-1, keepdim=True)
                
                # Create new input with the generated token
                next_token = next_token.unsqueeze(0)  # Add batch dimension if needed
                new_input = torch.cat([current_input, next_token], dim=1)
            
            # Decode the generated token
            token_text = self.tokenizer.decode(next_token.item())
            generated_tokens.append(next_token.item())
            
            # Create token score object
            token_score = TokenScore(
                token_id=next_token.item(),
                token_text=token_text,
                position=token_idx,
                similarity=0.0,  # Will be filled by monitor callback
                is_harmful=False,  # Will be filled by monitor callback
                category="unknown"  # Will be filled by monitor callback
            )
            
            # Call monitor callback if provided
            if monitor_callback is not None:
                try:
                    monitor_result = monitor_callback(outputs, new_input, token_score)
                    if monitor_result:
                        token_score.similarity = monitor_result.get('similarity', 0.0)
                        token_score.is_harmful = monitor_result.get('is_harmful', False)
                        token_score.category = monitor_result.get('category', 'unknown')
                except Exception as e:
                    # Continue generation even if monitoring fails
                    pass
            
            token_scores.append(token_score.to_dict())
            
            # Update current input for next iteration
            current_input = new_input
            
            # Check for early stopping conditions
            if next_token.item() == self.tokenizer.eos_token_id:
                break
        
        # Decode the full generated text
        if generated_tokens:
            generated_text = self.tokenizer.decode(generated_tokens, skip_special_tokens=True)
        else:
            generated_text = ""
        
        # Create result dictionary
        result = {
            "response": generated_text,
            "token_scores": token_scores,
            "prompt_length": prompt_length,
            "tokens_generated": len(generated_tokens),
            "format_type": self.format_type.value
        }
        
        return result

    def generate(self, prompt: str, layer_index: int, max_new_tokens: int = 50, enable_gradients: bool = False, 
                 nonsense_detector=None, nonsense_action: str = "regenerate", max_regeneration_attempts: int = 3,
                 latency_tracker=None, operation_name: str = "response_generation",
                 **generation_kwargs):
        """Generate text and extract activations from specified layer (legacy method)."""
        if self.hf_model is None:
            raise ValueError("No model loaded")
            
        # Format prompt
        formatted_prompt = self.format_prompt(prompt)
            
        # Track regeneration attempts for nonsense detection
        attempt = 0
        
        while attempt <= max_regeneration_attempts:
            try:
                # Use generation timing context if tracker provided
                if latency_tracker:
                    # Get prompt length for metrics
                    inputs = self.tokenizer(formatted_prompt, return_tensors="pt", padding=True, truncation=True)
                    prompt_length = inputs["input_ids"].shape[1]
                    
                    with latency_tracker.time_generation(operation_name, prompt_length) as gen_state:
                        generated_text, activations, token_count = self._generate_with_timing(
                            formatted_prompt, layer_index, max_new_tokens, enable_gradients, 
                            gen_state, **generation_kwargs
                        )
                else:
                    generated_text, activations, token_count = self._generate_with_timing(
                        formatted_prompt, layer_index, max_new_tokens, enable_gradients, 
                        None, **generation_kwargs
                    )
                
                # Check for nonsense if detector is provided
                if nonsense_detector is not None and nonsense_action != "flag":
                    result = nonsense_detector.detect_nonsense(generated_text)
                    
                    if result["is_nonsense"]:
                        if nonsense_action == "stop":
                            print(f"⚠️ Nonsense detected, stopping generation: {', '.join(result['issues'])}")
                            return "", None
                        elif nonsense_action == "regenerate" and attempt < max_regeneration_attempts:
                            print(f"⚠️ Nonsense detected (attempt {attempt + 1}), regenerating: {', '.join(result['issues'])}")
                            attempt += 1
                            # Increase temperature slightly for next attempt
                            generation_kwargs["temperature"] = generation_kwargs.get("temperature", 0.7) + 0.1
                            continue
                        else:
                            print(f"⚠️ Max regeneration attempts reached, returning potentially nonsensical response")
                
                # Add nonsense detection results to output if flagging
                if nonsense_detector is not None and nonsense_action == "flag":
                    result = nonsense_detector.detect_nonsense(generated_text)
                    if result["is_nonsense"]:
                        generated_text = f"[FLAGGED: {', '.join(result['issues'])}] {generated_text}"
                
                return generated_text, activations
                
            except Exception as e:
                if attempt < max_regeneration_attempts:
                    print(f"⚠️ Generation error (attempt {attempt + 1}), retrying: {e}")
                    attempt += 1
                    continue
                else:
                    raise e
        
        # If we get here, all attempts failed
        return "", None

    def _generate_with_timing(self, formatted_prompt: str, layer_index: int, max_new_tokens: int, 
                             enable_gradients: bool, gen_state=None, **generation_kwargs):
        """Helper method to handle generation with optional timing tracking."""
        import time
        
        # Prepare inputs
        inputs = self.tokenizer(formatted_prompt, return_tensors="pt", padding=True, truncation=True)
        inputs = {k: v.to(self.hf_model.device) for k, v in inputs.items()}
        input_length = inputs["input_ids"].shape[1]
        
        generation_config = {
            "max_new_tokens": max_new_tokens,
            "do_sample": True,
            "temperature": 0.7,
            "pad_token_id": self.tokenizer.pad_token_id,
            "output_hidden_states": True,
            "return_dict_in_generate": True,
            **generation_kwargs
        }
        
        # For TTFT tracking, we need to implement streaming-like generation
        if gen_state:
            # Use a simple approach: measure time to generate just the first token, then the rest
            first_token_config = generation_config.copy()
            first_token_config["max_new_tokens"] = 1
            
            # Generate first token
            if enable_gradients:
                first_outputs = self.hf_model.generate(**inputs, **first_token_config)
            else:
                with torch.inference_mode():
                    first_outputs = self.hf_model.generate(**inputs, **first_token_config)
            
            # Mark first token time
            gen_state["mark_first_token"]()
            
            # Continue with remaining tokens if max_new_tokens > 1
            if max_new_tokens > 1:
                # Update inputs with the first generated token
                new_inputs = {"input_ids": first_outputs.sequences, "attention_mask": torch.ones_like(first_outputs.sequences)}
                remaining_config = generation_config.copy()
                remaining_config["max_new_tokens"] = max_new_tokens - 1
                
                if enable_gradients:
                    outputs = self.hf_model.generate(**new_inputs, **remaining_config)
                else:
                    with torch.inference_mode():
                        outputs = self.hf_model.generate(**new_inputs, **remaining_config)
                
                # Combine sequences
                generated_tokens = outputs.sequences[0][input_length:]
            else:
                outputs = first_outputs
                generated_tokens = first_outputs.sequences[0][input_length:]
        else:
            # Standard generation without TTFT tracking
            if enable_gradients:
                outputs = self.hf_model.generate(**inputs, **generation_config)
            else:
                with torch.inference_mode():
                    outputs = self.hf_model.generate(**inputs, **generation_config)
            
            generated_tokens = outputs.sequences[0][input_length:]
        
        # Extract generated text
        generated_text = self.tokenizer.decode(generated_tokens, skip_special_tokens=True)
        token_count = len(generated_tokens)
        
        # Update token count for timing tracker
        if gen_state:
            gen_state["update_tokens"](token_count)
        
        # Extract activations from the specified layer
        if hasattr(outputs, 'hidden_states') and outputs.hidden_states:
            # Get hidden states from the last generation step
            last_hidden_states = outputs.hidden_states[-1]  # Last generation step
            if isinstance(last_hidden_states, tuple) and len(last_hidden_states) > layer_index:
                layer_activations = last_hidden_states[layer_index]
                # Get the last token's activations
                activations = layer_activations[0, -1, :].detach()
            else:
                activations = None
        else:
            activations = None
                    
        return generated_text, activations, token_count

    def extract_activations(self, text: str, layer: 'Layer'):
        """Extract activations from the specified layer for given text."""
        if self.hf_model is None:
            raise ValueError("No model loaded")
        
        # Format the text
        formatted_text = self.format_prompt(text)
            
        inputs = self.tokenizer(formatted_text, return_tensors="pt", padding=True, truncation=True)
        inputs = {k: v.to(self.hf_model.device) for k, v in inputs.items()}
        
        with torch.no_grad():
            outputs = self.hf_model(**inputs, output_hidden_states=True)
            
        # Get hidden states for the specified layer
        if layer.index + 1 < len(outputs.hidden_states):
            layer_hidden_states = outputs.hidden_states[layer.index + 1]  # +1 because [0] is embeddings
            
            # Extract last token's activations to ensure consistent shape
            # Shape: [batch_size, sequence_length, hidden_dim] -> [batch_size, hidden_dim]
            last_token_activations = layer_hidden_states[:, -1, :]
            
            return last_token_activations
        else:
            raise ValueError(f"Layer {layer.index} not found in model with {len(outputs.hidden_states)} layers")

    def prepare_activations(self, text: str) -> Dict[str, Any]:
        """
        Prepare activations for monitoring by formatting text and running through model.
        
        Args:
            text: Text to process
            
        Returns:
            Dictionary with model outputs and formatted inputs
        """
        # Format the prompt properly
        formatted_text = self.format_prompt(text)
        
        # Tokenize input
        inputs = self.tokenizer(formatted_text, return_tensors="pt", padding=True, truncation=True)
        inputs = {k: v.to(self.hf_model.device) for k, v in inputs.items()}
        
        # Run through model to get outputs
        with torch.no_grad():
            outputs = self.hf_model(inputs.input_ids, output_hidden_states=True)
        
        return {
            'outputs': outputs,
            'inputs': inputs,
            'formatted_text': formatted_text
        }

    # Parameter optimization functionality
    def optimize_parameters(
        self,
        harmful_texts: List[str],
        harmless_texts: List[str],
        layer_range: Tuple[int, int] = (10, 20),
        steering_types: List[str] = None,
        threshold_range: Tuple[float, float] = (0.3, 0.8),
        num_threshold_steps: int = 6
    ) -> Dict[str, Any]:
        """
        Optimize model parameters using grid search.
        
        Args:
            harmful_texts: List of harmful examples
            harmless_texts: List of harmless examples
            layer_range: Range of layers to test (start, end)
            steering_types: List of steering types to test
            threshold_range: Range of thresholds to test
            num_threshold_steps: Number of threshold values to test
            
        Returns:
            Optimization results with best parameters
        """
        from .steering import SteeringMethod, SteeringType
        # ContrastivePairSet import moved to top of file
        from .layer import Layer
        
        if steering_types is None:
            steering_types = ["logistic", "mlp"]
        
        # Create phrase pairs
        phrase_pairs = []
        min_len = min(len(harmful_texts), len(harmless_texts))
        
        for i in range(min_len):
            phrase_pairs.append({
                "harmful": harmful_texts[i],
                "harmless": harmless_texts[i]
            })
        
        # Split into train/test
        split_idx = int(0.8 * len(phrase_pairs))
        train_pairs = phrase_pairs[:split_idx]
        test_pairs = phrase_pairs[split_idx:]
        
        best_score = 0.0
        best_params = {}
        all_results = []
        
        # Grid search
        layers_to_test = list(range(layer_range[0], layer_range[1] + 1))
        thresholds_to_test = [
            threshold_range[0] + i * (threshold_range[1] - threshold_range[0]) / (num_threshold_steps - 1)
            for i in range(num_threshold_steps)
        ]
        
        total_combinations = len(layers_to_test) * len(steering_types) * len(thresholds_to_test)
        
        for layer_idx in layers_to_test:
            for steering_type in steering_types:
                for threshold in thresholds_to_test:
                    try:
                        # Create and train steering method
                        steering_method = SteeringMethod(
                            method_type=SteeringType(steering_type),
                            device=self.device
                        )
                        
                        # Create training pair set
                        train_pair_set = ContrastivePairSet.from_phrase_pairs(
                            name=f"train_layer_{layer_idx}",
                            phrase_pairs=train_pairs,
                            task_type="parameter_optimization"
                        )
                        
                        layer_obj = Layer(index=layer_idx, type="transformer")
                        
                        # Train
                        train_results = train_pair_set.train_classifier(
                            steering_method.classifier,
                            layer_obj
                        )
                        
                        # Test
                        test_pair_set = ContrastivePairSet.from_phrase_pairs(
                            name=f"test_layer_{layer_idx}",
                            phrase_pairs=test_pairs,
                            task_type="parameter_optimization"
                        )
                        
                        test_results = test_pair_set.evaluate_with_vectors(
                            steering_method,
                            layer_obj
                        )
                        
                        # Get score (use accuracy or F1)
                        score = test_results.get("accuracy", 0.0)
                        
                        result = {
                            "layer": layer_idx,
                            "steering_type": steering_type,
                            "threshold": threshold,
                            "score": score,
                            "train_results": train_results,
                            "test_results": test_results
                        }
                        
                        all_results.append(result)
                        
                        # Update best if better
                        if score > best_score:
                            best_score = score
                            best_params = {
                                "layer": layer_idx,
                                "steering_type": steering_type,
                                "threshold": threshold,
                                "score": score
                            }
                        
                    except Exception as e:
                        # Continue with next combination
                        pass
        
        return {
            "best_parameters": best_params,
            "best_score": best_score,
            "all_results": all_results,
            "total_combinations_tested": len(all_results)
        }
    
    def optimize_layer_selection(
        self,
        harmful_texts: List[str],
        harmless_texts: List[str],
        layer_range: Tuple[int, int] = (5, 25),
        steering_type: str = "logistic"
    ) -> Dict[str, Any]:
        """
        Optimize layer selection specifically.
        
        Args:
            harmful_texts: List of harmful examples
            harmless_texts: List of harmless examples
            layer_range: Range of layers to test
            steering_type: Steering type to use
            
        Returns:
            Layer optimization results
        """
        return self.optimize_parameters(
            harmful_texts=harmful_texts,
            harmless_texts=harmless_texts,
            layer_range=layer_range,
            steering_types=[steering_type],
            threshold_range=(0.5, 0.5),  # Fixed threshold
            num_threshold_steps=1
        )
    
    def optimize_threshold(
        self,
        harmful_texts: List[str],
        harmless_texts: List[str],
        layer: int = 15,
        steering_type: str = "logistic",
        threshold_range: Tuple[float, float] = (0.1, 0.9),
        num_steps: int = 10
    ) -> Dict[str, Any]:
        """
        Optimize threshold specifically.
        
        Args:
            harmful_texts: List of harmful examples
            harmless_texts: List of harmless examples
            layer: Fixed layer to use
            steering_type: Fixed steering type to use
            threshold_range: Range of thresholds to test
            num_steps: Number of threshold values to test
            
        Returns:
            Threshold optimization results
        """
        return self.optimize_parameters(
            harmful_texts=harmful_texts,
            harmless_texts=harmless_texts,
            layer_range=(layer, layer),  # Fixed layer
            steering_types=[steering_type],
            threshold_range=threshold_range,
            num_threshold_steps=num_steps
        )
    
    @staticmethod
    def get_available_tasks() -> List[str]:
        """
        Get list of all available tasks.
        
        Returns:
            List of available task names
        """
        return _get_available_tasks()
    
    @staticmethod
    def is_valid_task(task_name: str) -> bool:
        """
        Check if a task name is valid, including fuzzy matching.
        
        Args:
            task_name: Name of the task to check
            
        Returns:
            True if task is valid, False otherwise
        """
        return _is_valid_task(task_name)
    
    def load_lm_eval_task(self, task_name: str, shots: int = 0, limit: Optional[int] = None):
        """
        Load a task from lm-evaluation-harness with dynamic task name resolution.
        
        Args:
            task_name: Name of the task
            shots: Number of few-shot examples
            limit: Optional limit on number of documents
            
        Returns:
            Task object from lm_eval
        """
        # Check if it's LiveCodeBench
        if task_name == "livecodebench":
            from .tasks.livecodebench_task import LiveCodeBenchTask
            return LiveCodeBenchTask(release_version="release_v1", limit=limit)
            
        # Check if it's a BigCode task
        try:
            from .bigcode_integration import is_bigcode_task, load_bigcode_task
            
            if is_bigcode_task(task_name):
                # Load from BigCode
                return load_bigcode_task(task_name, limit=limit)
        except ImportError:
            # BigCode not available, continue with lm-eval
            pass
            
        # Default to lm-eval
        if _task_manager is None:
            raise RuntimeError("Task management system not available")
        
        return _task_manager.load_task(task_name, shots=shots, limit=limit)
    
    def _resolve_task_name(self, task_name: str) -> str:
        """Dynamically resolve a task name to an available task."""
        return _resolve_task_name(task_name)
    
    def _calculate_task_name_similarity(self, name1: str, name2: str) -> float:
        """Calculate similarity between two task names."""
        # Exact match
        if name1 == name2:
            return 1.0
        
        # Substring matching
        if name1 in name2 or name2 in name1:
            longer = max(name1, name2, key=len)
            shorter = min(name1, name2, key=len)
            return len(shorter) / len(longer)
        
        # Token-based similarity
        tokens1 = set(name1.replace('_', ' ').replace('-', ' ').split())
        tokens2 = set(name2.replace('_', ' ').replace('-', ' ').split())
        
        if not tokens1 or not tokens2:
            return 0.0
        
        intersection = len(tokens1 & tokens2)
        union = len(tokens1 | tokens2)
        
        return intersection / union if union > 0 else 0.0
    
    def split_task_data(self, task_data, split_ratio: float = 0.8, random_seed: int = 42):
        """
        Split task data into training and testing sets.
        
        Args:
            task_data: Task object from lm_eval
            split_ratio: Proportion for training set
            random_seed: Random seed for reproducibility
            
        Returns:
            Tuple of (train_docs, test_docs)
        """
        if _task_manager is None:
            raise RuntimeError("Task management system not available")
        
        return _task_manager.split_task_data(task_data, split_ratio=split_ratio, random_seed=random_seed)
    
    def prepare_prompts_from_docs(self, task, docs: List[Dict[str, Any]]) -> List[str]:
        """
        Prepare prompts from task documents.
        
        Args:
            task: Task object from lm_eval
            docs: List of documents
            
        Returns:
            List of formatted prompts
        """
        if _task_manager is None:
            raise RuntimeError("Task management system not available")
        
        return _task_manager.prepare_prompts_from_docs(task, docs)
    
    def get_reference_answers(self, task, docs: List[Dict[str, Any]]) -> List[str]:
        """
        Extract reference answers from task documents.
        
        Args:
            task: Task object from lm_eval
            docs: List of documents
            
        Returns:
            List of reference answers
        """
        if _task_manager is None:
            raise RuntimeError("Task management system not available")
        
        return _task_manager.get_reference_answers(task, docs)


class ModelParameterOptimizer:
    """
    Parameter optimizer integrated into the model primitive.
    """
    
    def __init__(
        self,
        model_name: str,
        device: Optional[str] = None
    ):
        """Initialize parameter optimizer."""
        self.model = Model(name=model_name, device=device)
        self.optimization_history = []
    
    def optimize_parameters(
        self,
        harmful_texts: List[str],
        harmless_texts: List[str],
        **kwargs
    ) -> Dict[str, Any]:
        """Optimize parameters using the model's optimization functionality."""
        result = self.model.optimize_parameters(harmful_texts, harmless_texts, **kwargs)
        self.optimization_history.append(result)
        return result
    
    def optimize_layer_selection(self, *args, **kwargs) -> Dict[str, Any]:
        """Optimize layer selection."""
        result = self.model.optimize_layer_selection(*args, **kwargs)
        self.optimization_history.append(result)
        return result
    
    def optimize_threshold(self, *args, **kwargs) -> Dict[str, Any]:
        """Optimize threshold."""
        result = self.model.optimize_threshold(*args, **kwargs)
        self.optimization_history.append(result)
        return result
    
    def get_optimization_history(self) -> List[Dict[str, Any]]:
        """Get history of optimization runs."""
        return self.optimization_history
    
    def clear_history(self) -> None:
        """Clear optimization history."""
        self.optimization_history.clear()


class ActivationHooks:
    """
    Activation hooks functionality integrated into the model primitive.
    """
    
    def __init__(
        self,
        model: torch.nn.Module,
        layers: List[int],
        token_strategy: str = "last"
    ):
        """Initialize activation hooks for a model."""
        self.model = model
        self.layers = layers
        self.token_strategy = token_strategy
        
        # Determine model type
        self.model_type = self._detect_model_type(model)
        
        # Initialize hook storage
        self.hooks = {}
        self.activations = {}
        self.layer_activations = {}
        self.active_layers = set()
        self.last_token_id = None
        self.last_token_position = None
        
        # Set up monitoring on specified layers
        self.setup_hooks(layers)
    
    def _detect_model_type(self, model) -> str:
        """Detect model type."""
        model_config = getattr(model, "config", None)
        model_name = getattr(model_config, "_name_or_path", "unknown").lower()
        
        if hasattr(model, "get_input_embeddings"):
            if "llama" in model_name:
                return "llama"
            elif "mistral" in model_name:
                return "mistral"
            elif "mpt" in model_name:
                return "mpt"
        
        return "generic"
    
    def _get_layer_name(self, model_type: str, layer_idx: int) -> str:
        """Get the layer name for a given model type and layer index."""
        if model_type == "llama":
            return f"model.layers.{layer_idx}"
        elif model_type == "mistral":
            return f"model.layers.{layer_idx}"
        elif model_type == "mpt":
            return f"transformer.blocks.{layer_idx}"
        elif model_type == "gpt2":
            return f"transformer.h.{layer_idx}"
        else:
            # Check user config for custom models
            if user_model_configs.has_config(self.name):
                layer_info = user_model_configs.get_layer_access_info(self.name)
                if layer_info and layer_info.get("layer_path_template"):
                    return layer_info["layer_path_template"].format(idx=layer_idx)
            
            # No config found - this shouldn't happen since we prompt during init
            # but just in case, prompt again
            print(f"\n⚠️  Model '{self.name}' layer configuration is missing.")
            response = input("Would you like to configure it now? (y/n): ").strip().lower()
            
            if response == 'y':
                config = user_model_configs.prompt_and_save_config(self.name)
                layer_info = user_model_configs.get_layer_access_info(self.name)
                if layer_info and layer_info.get("layer_path_template"):
                    return layer_info["layer_path_template"].format(idx=layer_idx)
            
            raise ValueError(
                f"Model '{self.name}' requires layer configuration.\n"
                f"Run 'wisent-guard configure-model {self.name}' to set it up."
            )
    
    def _get_module_by_name(self, name: str) -> torch.nn.Module:
        """Retrieve a module from the model by its name."""
        module = self.model
        for part in name.split('.'):
            if part.isdigit():
                module = module[int(part)]
            else:
                module = getattr(module, part)
        return module
    
    def _activation_hook(self, layer_idx: int):
        """Create a hook function for the specified layer."""
        def hook(module, input, output):
            if layer_idx in self.active_layers:
                # Get the output hidden states
                if isinstance(output, tuple):
                    hidden_states = output[0]
                else:
                    hidden_states = output
                
                if isinstance(hidden_states, torch.Tensor):
                    device = hidden_states.device
                    
                    # Get last token's activations
                    last_token_idx = hidden_states.shape[1] - 1
                    
                    # Store activations - preserve gradients if needed for steering
                    # Check if we're in a gradient-enabled context (for steering methods like K-steering)
                    if torch.is_grad_enabled() and hidden_states.requires_grad:
                        # Preserve computational graph for gradient-based steering
                        self.layer_activations[layer_idx] = hidden_states[:, last_token_idx, :].clone().to(device)
                    else:
                        # Standard extraction - detach for memory efficiency
                        self.layer_activations[layer_idx] = hidden_states[:, last_token_idx, :].detach().clone().to(device)
                    
                    self.last_token_position = last_token_idx
                    
                    # Try to get the token ID if available
                    if hasattr(module, '_last_input_ids'):
                        input_ids = module._last_input_ids[0]
                        if last_token_idx < len(input_ids):
                            try:
                                self.last_token_id = input_ids[last_token_idx].item()
                            except (RuntimeError, ValueError):
                                self.last_token_id = None
        return hook
    
    def register_hooks(self, layers: List[int]) -> None:
        """Register activation hooks for the specified layers."""
        self.active_layers = set(layers)
        
        # Clear existing hooks
        self.remove_hooks()
        
        # Register new hooks
        for layer_idx in layers:
            layer_name = self._get_layer_name(self.model_type, layer_idx)
            
            try:
                module = self._get_module_by_name(layer_name)
                
                # Add a pre-hook to capture input_ids
                def forward_pre_hook(module, args):
                    if len(args) > 0 and isinstance(args[0], torch.Tensor):
                        module._last_input_ids = args[0]
                    return args
                
                module.register_forward_pre_hook(forward_pre_hook)
                
                # Add the main activation hook
                hook = module.register_forward_hook(self._activation_hook(layer_idx))
                self.hooks[layer_idx] = hook
                
            except Exception:
                # Continue with other layers if one fails
                pass
    
    def remove_hooks(self) -> None:
        """Remove all registered hooks."""
        if self.hooks:
            for layer_idx, hook in self.hooks.items():
                hook.remove()
            self.hooks = {}
    
    def setup_hooks(self, layers: List[int]):
        """Set up hooks for specified layers."""
        self.register_hooks(layers)
    
    def add_steering_hooks(self, layers: List[int], steering_function) -> List:
        """Add steering hooks to specified layers.
        
        Args:
            layers: List of layer indices to add hooks to
            steering_function: Function that takes (module, input, output) and returns modified output
            
        Returns:
            List of registered hooks that can be removed later
        """
        hooks = []
        
        for layer_idx in layers:
            layer_name = self._get_layer_name(self.model_type, layer_idx)
            
            try:
                module = self._get_module_by_name(layer_name)
                hook = module.register_forward_hook(steering_function)
                hooks.append(hook)
            except Exception as e:
                # Log error but continue with other layers
                print(f"Warning: Could not add steering hook to layer {layer_idx}: {e}")
                
        return hooks
    
    def has_activations(self) -> bool:
        """Check if we have collected any activations."""
        return bool(self.layer_activations)
    
    def get_activations(self) -> Dict[int, torch.Tensor]:
        """Get all activations for registered layers."""
        return self.layer_activations
    
    def get_last_token_info(self) -> Dict[str, Any]:
        """Get information about the last token processed."""
        return {
            "token_id": self.last_token_id,
            "position": self.last_token_position
        }
    
    def reset(self):
        """Reset stored activations."""
        self.activations = {}
        self.layer_activations = {}
    
    def clear_activations(self):
        """Clear all stored activations."""
        self.reset()


# LM Evaluation Integration - now handled by task management system
# Import from the dedicated task management system
try:
    from .agent.diagnose.tasks import (
        get_available_tasks as _get_available_tasks,
        is_valid_task as _is_valid_task,
        resolve_task_name as _resolve_task_name,
        TaskManager,
        load_docs
    )
    # Initialize global task manager for Model class methods
    _task_manager = TaskManager()
except ImportError:
    # Fallback if task management system not available
    _task_manager = None
    def _get_available_tasks():
        return []
    def _is_valid_task(task_name):
        return False
    def _resolve_task_name(task_name):
        raise ValueError(f"Task management system not available")
    def load_docs(task, limit=None):
        raise ValueError(f"Task management system not available")


"""
Steering methods package for wisent-guard.

This package contains implementations of various steering methods:
- CAA: Contrastive Activation Addition
- HPR: Householder Pseudo-Rotation  
- DAC: Dynamic Activation Composition
- BiPO: Bi-directional Preference Optimization
- K-Steering: Multi-directional steering using gradient-based optimization
"""

from .base import SteeringMethod
from .caa import CAA
from .hpr import HPR
from .dac import DAC
from .bipo import BiPO
from .k_steering import KSteering

__all__ = [
    'SteeringMethod',
    'CAA',
    'HPR', 
    'DAC',
    'BiPO',
    'KSteering'
] 


class Layer:
    def __init__(self, index, type=None):
        self.index = index
        self.type = type 

    def get_activations(self, model, prompt, **kwargs):
        _, activations = model.generate(prompt, self.index, **kwargs)
        return activations 